import json
from typing import Dict, Any, List, Optional
from flo_ai.models.base_agent import BaseAgent, AgentType, ReasoningPattern
from flo_ai.llm.base_llm import BaseLLM
from flo_ai.models.chat_message import (
    AssistantMessage,
    BaseMessage,
    MessageType,
    UserMessage,
    TextMessageContent,
    FunctionMessage,
)
from flo_ai.tool.base_tool import Tool, ToolExecutionError
from flo_ai.models.agent_error import AgentError
from flo_ai.utils.logger import logger
from flo_ai.utils.variable_extractor import (
    extract_variables_from_inputs,
    extract_agent_variables,
    validate_multi_agent_variables,
    resolve_variables,
)
from flo_ai.telemetry.instrumentation import (
    trace_agent_execution,
    agent_metrics,
)
from flo_ai.telemetry import get_tracer


class Agent(BaseAgent):
    def __init__(
        self,
        name: str,
        system_prompt: str | AssistantMessage,
        llm: BaseLLM,
        tools: Optional[List[Tool]] = None,
        max_retries: int = 0,
        max_tool_calls: int = 5,
        reasoning_pattern: ReasoningPattern = ReasoningPattern.DIRECT,
        output_schema: Optional[Dict[str, Any]] = None,
        role: Optional[str] = None,
        act_as: Optional[str] = MessageType.ASSISTANT,
        input_filter: Optional[List[str]] = None,
    ):
        # Determine agent type based on tools
        agent_type = AgentType.TOOL_USING if tools else AgentType.CONVERSATIONAL

        # Enhance system prompt with role if provided
        enhanced_prompt = system_prompt
        if role:
            if isinstance(system_prompt, str):
                enhanced_prompt = f'You are {role}. {system_prompt}'
            elif isinstance(system_prompt, AssistantMessage):
                enhanced_prompt = f'You are {role}. {system_prompt.content}'

        super().__init__(
            name=name,
            system_prompt=enhanced_prompt,
            agent_type=agent_type,
            llm=llm,
            max_retries=max_retries,
            max_tool_calls=max_tool_calls,
        )
        self.tools = tools or []
        self.tools_dict = {tool.name: tool for tool in self.tools}
        self.reasoning_pattern = reasoning_pattern
        self.output_schema = output_schema
        self.role = role
        self.act_as = act_as
        self.input_filter: Optional[List[str]] = input_filter

    @trace_agent_execution()
    async def run(
        self,
        inputs: List[BaseMessage] | str,
        variables: Optional[Dict[str, Any]] = None,
    ) -> str:
        variables = variables or {}
        if isinstance(inputs, str):
            inputs = [UserMessage(TextMessageContent(text=inputs))]

        # Perform runtime variable validation if not already resolved (single agent usage)
        if not self.resolved_variables:
            # Extract variables from inputs and system prompt
            input_variables = extract_variables_from_inputs(inputs)
            agent_variables = extract_agent_variables(self)
            all_required_variables = input_variables.union(agent_variables)

            # Validate that all required variables are provided
            if all_required_variables:
                agents_variables = {self.name: all_required_variables}
                validate_multi_agent_variables(agents_variables, variables)

            # Resolve variables and mark as resolved
            self.system_prompt = resolve_variables(self.system_prompt, variables)

            # Process inputs and resolve variables in string inputs
            for input in inputs:
                if isinstance(input, BaseMessage):
                    # checking whether the TextMessageContent is resolved
                    if isinstance(input.content, TextMessageContent) and variables:
                        input.content.text = resolve_variables(
                            input.content.text, variables
                        )
                    self.add_to_history(input)
                else:
                    raise ValueError(f'Invalid input type: {type(input)}')
            # after resolving agent system prompts and inputs, mark variables as resolved
            self.resolved_variables = True

        else:
            # Variables already resolved, process inputs without variable resolution
            for input in inputs:
                if isinstance(input, BaseMessage):
                    # Handle InputMessage - check content type
                    self.add_to_history(input)
                else:
                    raise ValueError(f'Invalid input type: {type(input)}')

        retry_count = 0

        # If no tools, act as conversational agent
        if not self.tools:
            return await self._run_conversational(retry_count, variables)

        # Otherwise, run as tool agent
        return await self._run_with_tools(retry_count, variables)

    async def _run_conversational(
        self, retry_count: int, variables: Optional[Dict[str, Any]] = None
    ) -> str:
        """Run as a conversational agent when no tools are provided"""
        variables = variables or {}

        while retry_count <= self.max_retries:
            try:
                # Resolve variables in system prompt
                system_content = (
                    self._get_cot_prompt(variables)
                    if self.reasoning_pattern == ReasoningPattern.COT
                    else resolve_variables(self.system_prompt, variables)
                )
                system_message = AssistantMessage(
                    role=MessageType.SYSTEM,
                    content=TextMessageContent(text=system_content),
                )
                self.add_to_history(system_message)
                messages = await self._get_message_history(variables)

                logger.debug(f'Sending messages to LLM: {messages}')
                response = await self.llm.generate(
                    messages, output_schema=self.output_schema
                )
                logger.debug(f'Raw LLM Response: {response}')

                assistant_message = self.llm.get_message_content(response)
                logger.debug(f'Extracted message: {assistant_message}')

                # Ensure act_as is not None (default to 'assistant' if missing)
                role = self.act_as if self.act_as is not None else MessageType.ASSISTANT

                if assistant_message:
                    self.add_to_history(
                        AssistantMessage(role=role, content=assistant_message)
                    )
                else:
                    possible_tool_message = await self.llm.get_function_call(response)
                    if possible_tool_message:
                        self.add_to_history(
                            AssistantMessage(
                                role=role, content=possible_tool_message['arguments']
                            )
                        )
                    else:
                        logger.debug('Warning: No message content found in response')
                        self.add_to_history(
                            AssistantMessage(
                                role=role,
                                content='No message content found in response',
                            )
                        )

                return self.conversation_history

            except Exception as e:
                retry_count += 1
                context = {
                    'conversation_history': self.conversation_history,
                    'attempt': retry_count,
                }

                should_retry, analysis = await self.handle_error(e, context)

                if should_retry and retry_count <= self.max_retries:
                    self.add_to_history(
                        AssistantMessage(
                            content=f'Error occurred. Analysis: {analysis}'
                        )
                    )
                    continue
                else:
                    raise AgentError(
                        f'Failed after {retry_count} attempts. Last error: {analysis}',
                        original_error=e,
                    )

    async def _run_with_tools(
        self, retry_count: int = 0, variables: Optional[Dict[str, Any]] = None
    ) -> str:
        """Run as a tool-using agent when tools are provided"""
        variables = variables or {}
        print('running with tools')

        while retry_count <= self.max_retries:
            try:
                # Resolve variables in system prompt based on reasoning pattern
                if self.reasoning_pattern == ReasoningPattern.REACT:
                    system_content = self._get_react_prompt(variables)
                elif self.reasoning_pattern == ReasoningPattern.COT:
                    system_content = self._get_cot_prompt(variables)
                else:
                    system_content = resolve_variables(self.system_prompt, variables)

                system_message = AssistantMessage(
                    role=MessageType.SYSTEM, content=system_content
                )
                self.add_to_history(system_message)
                messages = await self._get_message_history(variables)

                # Keep executing tools until we get a final answer
                tool_call_count = 0
                while tool_call_count < self.max_tool_calls:
                    formatted_tools = self.llm.format_tools_for_llm(self.tools)
                    response = await self.llm.generate(
                        messages,
                        functions=formatted_tools,
                        output_schema=self.output_schema,
                    )

                    # Handle ReACT and CoT patterns
                    function_call = await self.llm.get_function_call(response)

                    # If no function call, check if this is truly a final answer
                    if not function_call:
                        assistant_message = self.llm.get_message_content(response)
                        if assistant_message:
                            # Check if this is a final answer or just intermediate reasoning
                            is_final = await self._is_final_answer(
                                assistant_message, tool_call_count, messages
                            )
                            if is_final:
                                # Ensure act_as is not None (default to 'assistant' if missing)
                                role = (
                                    self.act_as
                                    if self.act_as is not None
                                    else MessageType.ASSISTANT
                                )
                                self.add_to_history(
                                    AssistantMessage(
                                        role=role, content=assistant_message
                                    )
                                )
                                return self.conversation_history
                            else:
                                # This is intermediate reasoning, add to context and continue
                                msg_preview = (
                                    assistant_message[:100]
                                    if len(assistant_message) > 100
                                    else assistant_message
                                )
                                logger.debug(
                                    f'Detected intermediate reasoning (not final answer): {msg_preview}...'
                                )
                                # Ensure act_as is not None (default to 'assistant' if missing)
                                role = (
                                    self.act_as
                                    if self.act_as is not None
                                    else MessageType.ASSISTANT
                                )
                                self.add_to_history(
                                    AssistantMessage(
                                        role=role, content=assistant_message
                                    )
                                )
                                self.add_to_history(
                                    UserMessage(
                                        content='Based on your reasoning, please proceed with the necessary tool calls to complete the task.',
                                    )
                                )
                                continue
                        break

                    # If there's a function call, add the assistant's response
                    # LLM-specific implementations handle special formatting (e.g., Claude's raw_content)
                    assistant_message_content = (
                        self.llm.get_assistant_message_for_tool_call(response)
                    )
                    if assistant_message_content:
                        # LLM returned special formatting (e.g., Claude's raw_content)
                        messages.append(
                            {
                                'role': self.act_as,
                                'content': assistant_message_content,
                            }
                        )
                    else:
                        # Use default text content extraction
                        assistant_text = self.llm.get_message_content(response)
                        if assistant_text:
                            messages.append(
                                {
                                    'role': self.act_as,
                                    'content': assistant_text,
                                }
                            )

                    # Execute the tool
                    try:
                        function_name = function_call['name']
                        # Get tool_use_id if available (LLM-specific, e.g., Claude)
                        tool_use_id = self.llm.get_tool_use_id(function_call)
                        if isinstance(function_call['arguments'], str):
                            function_args = json.loads(function_call['arguments'])
                        else:
                            function_args = function_call['arguments']

                        tool = self.tools_dict[function_name]

                        # Track tool execution with telemetry
                        tracer = get_tracer()

                        if tracer:
                            with tracer.start_as_current_span(
                                f'agent.tool.{function_name}',
                                attributes={
                                    'tool.name': function_name,
                                    'agent.name': self.name,
                                },
                            ) as tool_span:
                                function_response = await tool.run(
                                    inputs=[], variables=None, **function_args
                                )
                                tool_span.set_attribute(
                                    'tool.result.length', len(str(function_response))
                                )
                        else:
                            function_response = await tool.run(
                                inputs=[], variables=None, **function_args
                            )

                        agent_metrics.record_tool_call(
                            self.name, function_name, 'success'
                        )

                        tool_call_count += 1

                        # Add function call result to history using OpenAI's "function" role format
                        # According to OpenAI API: {"role": "function", "name": "<function-name>", "content": "<result>"}
                        self.add_to_history(
                            FunctionMessage(
                                content=str(
                                    'Here is the result of the tool call: \n'
                                    + str(function_response)
                                ),
                                name=function_name,
                            )
                        )

                        # Add the function response to messages for context
                        # LLM-specific implementations format the message appropriately
                        function_result_msg = self.llm.format_function_result_message(
                            function_name, str(function_response), tool_use_id
                        )
                        messages.append(function_result_msg)

                    except (json.JSONDecodeError, KeyError, ToolExecutionError) as e:
                        # Record tool call failure
                        agent_metrics.record_tool_call(
                            self.name, function_name, 'error'
                        )

                        retry_count += 1
                        context = {
                            'function_call': function_call,
                            'attempt': retry_count,
                        }
                        should_retry, analysis = await self.handle_error(e, context)
                        if should_retry and retry_count <= self.max_retries:
                            # Record retry
                            agent_metrics.record_retry(
                                self.name, 'tool_execution_error'
                            )

                            self.add_to_history(
                                AssistantMessage(
                                    content=f'Tool execution error: {analysis}'
                                )
                            )
                            continue
                        raise AgentError(
                            f'Tool execution failed: {analysis}', original_error=e
                        )

                # Generate final response if we've hit the tool call limit or exited the loop
                system_message = AssistantMessage(
                    role=MessageType.SYSTEM,
                    content=TextMessageContent(
                        text='Please provide a final answer based on all the tool results above.'
                    ),
                )
                self.add_to_history(system_message)
                messages = await self._get_message_history(variables)

                final_response = await self.llm.generate(
                    messages,
                    output_schema=self.output_schema,
                )

                assistant_message = self.llm.get_message_content(final_response)
                if assistant_message:
                    # Ensure act_as is not None (default to 'assistant' if missing)
                    role = (
                        self.act_as
                        if self.act_as is not None
                        else MessageType.ASSISTANT
                    )
                    self.add_to_history(AssistantMessage(content=assistant_message))
                    return self.conversation_history

                return f'The final result based on the tool executions is: {function_response}'

            except Exception as e:
                retry_count += 1
                context = {
                    'conversation_history': self.conversation_history,
                    'attempt': retry_count,
                }

                should_retry, analysis = await self.handle_error(e, context)
                if should_retry and retry_count <= self.max_retries:
                    # Record retry
                    agent_metrics.record_retry(self.name, 'execution_error')

                    self.add_to_history(
                        AssistantMessage(
                            content=f'Error occurred. Analysis: {analysis}'
                        )
                    )
                    continue

                raise AgentError(
                    f'Failed after {retry_count} attempts. Last error: {analysis}',
                    original_error=e,
                )

        raise AgentError(f'Failed after maximum {self.max_retries} attempts.')

    def _get_react_prompt(self, variables: Optional[Dict[str, Any]] = None) -> str:
        """Get system prompt modified for ReACT pattern"""
        variables = variables or {}

        tools_desc = '\n'.join(
            [f'- {tool.name}: {tool.description}' for tool in self.tools]
        )

        # Resolve variables in the base system prompt
        resolved_system_prompt = resolve_variables(self.system_prompt, variables)

        react_prompt = f"""{resolved_system_prompt}
            When solving tasks, follow this format:

            Thought: Analyze the situation and think about what to do
            Action: Use available tools in the format: tool_name(param1: "value1", param2: "value2")
            Observation: The result of the action
            ... (repeat Thought/Action/Observation if needed)
            Final Answer: [Your complete answer to the user's question]

            Available tools:
            {tools_desc}

            Remember to:
            1. Think carefully about what needs to be done
            2. Use tools when needed
            3. Make observations about tool results
            4. Conclude with a final answer when the task is complete

            IMPORTANT: When you have enough information to answer the user's question, you MUST prefix your response with "Final Answer:" to indicate completion."""

        return react_prompt

    def _get_cot_prompt(self, variables: Optional[Dict[str, Any]] = None) -> str:
        """Get system prompt modified for Chain of Thought pattern"""
        variables = variables or {}

        tools_desc = '\n'.join(
            [f'- {tool.name}: {tool.description}' for tool in self.tools]
        )

        # Resolve variables in the base system prompt
        resolved_system_prompt = resolve_variables(self.system_prompt, variables)

        cot_prompt = f"""{resolved_system_prompt}
            When solving tasks, follow this Chain of Thought reasoning format:

            Let me think through this step by step:
            1. First, I need to understand what is being asked...
            2. Then, I should consider what information or tools I need.... Use available tools in the format: tool_name(param1: "value1", param2: "value2")
            3. Next, I'll analyze the available options...
            4. Finally, I'll provide a well-reasoned answer...

            Available tools:
            {tools_desc}

            Remember to:
            1. Break down complex problems into smaller steps
            2. Think through each step logically
            3. Use tools when needed to gather information
            4. Provide clear reasoning for your conclusions
            5. End with a final, well-justified answer

            IMPORTANT: When you have gathered all necessary information and are ready to provide your complete answer, you MUST prefix your response with "Final Answer:" to indicate completion."""

        return cot_prompt

    async def _is_final_answer(
        self, message: str, tool_call_count: int, messages: List[Dict[str, Any]]
    ) -> bool:
        """
        Determine if a message is a final answer or intermediate reasoning.
        Uses structured token detection (like LangChain's ReAct) with LLM fallback.

        Approach inspired by LangChain/CrewAI:
        1. Primary: Check for explicit "Final Answer:" token
        2. Fallback: Use LLM-based classification for robustness
        """
        message_stripped = message.strip()
        message_lower = message_stripped.lower()

        # Primary Detection: Explicit "Final Answer:" token (ReAct pattern)
        # This is the most reliable method used by LangChain and similar frameworks
        if message_stripped.startswith('Final Answer:') or message_lower.startswith(
            'final answer:'
        ):
            logger.debug('Explicit "Final Answer:" token detected - this is final')
            return True

        # Check if "Final Answer:" appears anywhere in the response
        # (agent might add context before the token)
        if 'final answer:' in message_lower:
            logger.debug('"Final Answer:" token found in response - treating as final')
            return True

        # Secondary Detection: Use LLM-based analysis for cases without explicit tokens
        # This handles:
        # - Agents not following the format perfectly
        # - Direct mode (without ReAct/CoT patterns)
        # - Edge cases where the agent provides answer without token

        analysis_prompt = f"""You are a classifier that determines if an AI agent's response is a FINAL ANSWER or INTERMEDIATE REASONING.

Agent's Response:
"{message_stripped}"

Context:
- Tool calls executed so far: {tool_call_count}
- Total conversation turns: {len(messages)}

Classification Criteria:

FINAL ANSWER - The response is final if it:
✓ Directly answers the user's original question with concrete information
✓ Provides specific data, results, or conclusions
✓ Does not suggest or request additional actions
✓ Reads like a complete, standalone answer
✓ Contains synthesis of information already gathered

INTERMEDIATE REASONING - The response is intermediate if it:
✗ Describes plans or intentions for what to do next
✗ Expresses need to gather more information
✗ Contains thinking/reasoning WITHOUT providing the actual answer
✗ Poses questions or expresses uncertainty about next steps
✗ Mentions specific tools it wants to use

Examples of INTERMEDIATE:
- "I need to query the database schema first"
- "Let me check the table structure"
- "First, I should examine..."

Examples of FINAL:
- "Based on the query results, the table contains 1,245 records..."
- "The analysis shows that revenue increased by 23%..."
- "After examining the data, the answer is..."

Respond with EXACTLY one word: "FINAL" or "INTERMEDIATE"
"""

        try:
            analysis_messages = [
                {
                    'role': MessageType.SYSTEM,
                    'content': 'You are a precise classification system. Respond with only FINAL or INTERMEDIATE.',
                },
                {'role': MessageType.USER, 'content': analysis_prompt},
            ]
            analysis_response = await self.llm.generate(analysis_messages)
            analysis = self.llm.get_message_content(analysis_response).strip().upper()

            is_final = 'FINAL' in analysis
            msg_preview = (
                message_stripped[:80]
                if len(message_stripped) > 80
                else message_stripped
            )
            logger.debug(
                f'LLM classifier: "{analysis}" -> is_final={is_final} (message preview: "{msg_preview}...")'
            )
            return is_final

        except Exception as e:
            logger.warning(
                f'LLM classification failed: {e}. Defaulting to final=False to allow continuation.'
            )
            # Conservative default: treat as intermediate to avoid premature exit
            # This is safer as it allows the agent to continue rather than stopping too early
            return False



"""
Variable extraction utilities for the Flo AI framework.

This module provides functions to extract variable placeholders from text,
inputs, and agent configurations for runtime variable validation.
"""

import re
from typing import List, Set, Dict, Any

from flo_ai.models.chat_message import BaseMessage, TextMessageContent, AssistantMessage


def extract_variables_from_text(text: str | AssistantMessage) -> Set[str]:
    """Extract variable placeholders from text using <variable_name> pattern.

    Args:
        text: Text containing variable placeholders like <variable_name>

    Returns:
        Set of variable names found in the text

    Examples:
        >>> extract_variables_from_text("Hello <name>, your <role> is important")
        {'name', 'role'}
        >>> extract_variables_from_text("No variables here")
        set()
    """
    if not text:
        return set()
    if isinstance(text, AssistantMessage):
        text_str = text.content
    elif isinstance(text, str):
        text_str = text

    # Use regex to find all <variable_name> patterns
    # \w+ matches word characters (letters, digits, underscore)
    variable_pattern = r'<(\w+)>'
    matches = re.findall(variable_pattern, text_str)
    return set(matches)


def extract_variables_from_inputs(inputs: List[BaseMessage]) -> Set[str]:
    """Extract variable placeholders from a list of input messages.

    Args:
        inputs: List of input strings or ImageMessages

    Returns:
        Set of variable names found across all string inputs

    Note:
        ImageMessageContent objects are skipped as they don't contain variable placeholders
    """
    all_variables = set()
    for input_item in inputs:
        if isinstance(input_item, BaseMessage):
            if isinstance(input_item.content, TextMessageContent):
                variables = extract_variables_from_text(input_item.content.text)
                all_variables.update(variables)
        if isinstance(input_item, str):
            variables = extract_variables_from_text(input_item)
            all_variables.update(variables)
        # Skip ImageMessageContent objects as they don't contain text variables

    return all_variables


def extract_agent_variables(agent) -> Set[str]:
    """Extract variable placeholders from an agent's system prompt.

    Args:
        agent: Agent instance with a system_prompt attribute

    Returns:
        Set of variable names found in the agent's system prompt

    Note:
        This function avoids importing Agent to prevent circular imports
    """
    if not hasattr(agent, 'system_prompt'):
        return set()

    return extract_variables_from_text(agent.system_prompt)


def validate_variables(
    required_variables: Set[str], provided_variables: Dict[str, Any], context: str = ''
) -> None:
    """Validate that all required variables are provided.

    Args:
        required_variables: Set of variable names that are required
        provided_variables: Dictionary of variable name to value mappings
        context: Optional context string for error messages (e.g., agent name)

    Raises:
        ValueError: If any required variables are missing
    """
    missing = required_variables - set(provided_variables.keys())
    if missing:
        available = list(provided_variables.keys()) if provided_variables else []
        context_str = f' for {context}' if context else ''
        raise ValueError(
            f'Missing required variables{context_str}: {sorted(missing)}. '
            f'Available variables: {available}'
        )


def resolve_variables(
    text: str | BaseMessage | AssistantMessage, variables: Dict[str, Any]
) -> str | BaseMessage | AssistantMessage:
    """Replace <variable_name> patterns with actual values

    Args:
        text: Text containing variable placeholders like <variable_name>
        variables: Dictionary of variable name to value mappings

    Returns:
        Text with variables resolved

    Raises:
        ValueError: If a variable placeholder is found but not provided in variables
    """
    if not text or not variables:
        return text

    def replace_var(match):
        var_name = match.group(1)
        if var_name in variables:
            return str(variables[var_name])
        else:
            available = list(variables.keys())
            raise ValueError(
                f"Variable '{var_name}' referenced in text but not provided. "
                f'Available variables: {available}'
            )

    return re.sub(r'<(\w+)>', replace_var, text)


def validate_multi_agent_variables(
    agents_variables: Dict[str, Set[str]], provided_variables: Dict[str, Any]
) -> None:
    """Validate variables for multiple agents and provide detailed error messages.

    Args:
        agents_variables: Dictionary mapping agent names to their required variables
        provided_variables: Dictionary of variable name to value mappings

    Raises:
        ValueError: If any agents are missing required variables, with detailed breakdown
    """
    missing_by_agent = {}
    provided_keys = set(provided_variables.keys())

    for agent_name, required_vars in agents_variables.items():
        missing = required_vars - provided_keys
        if missing:
            missing_by_agent[agent_name] = sorted(missing)

    if missing_by_agent:
        error_msg = 'Missing required variables for agents:\n'
        for agent_name, missing_vars in missing_by_agent.items():
            error_msg += f"  - Agent '{agent_name}': {missing_vars}\n"
        error_msg += f'Provided variables: {sorted(provided_keys)}'
        raise ValueError(error_msg)



import logging
import os

log_level = os.environ.get('FLO_AI_LOG_LEVEL', 'INFO')
log_format = (
    '%(asctime)s | %(levelname)-8s | %(name)s | %(filename)s:%(lineno)d | %(message)s'
)

logger = logging.getLogger('flo_ai')
logger.setLevel(log_level)

# Prevent affecting the root logger
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

# Optional: stop logs from propagating to the root logger
logger.propagate = False



from .base_llm import BaseLLM
from .anthropic_llm import Anthropic
from .openai_llm import OpenAI
from .ollama_llm import OllamaLLM
from .gemini_llm import Gemini
from .openai_vllm import OpenAIVLLM
from .vertexai_llm import VertexAI
from .rootflo_llm import RootFloLLM

__all__ = [
    'BaseLLM',
    'Anthropic',
    'OpenAI',
    'OllamaLLM',
    'Gemini',
    'OpenAIVLLM',
    'VertexAI',
    'RootFloLLM',
]



"""
LLM Factory - Centralized LLM creation from configuration.

This module provides a unified factory function for creating LLM instances
from configuration dictionaries, supporting all providers in the flo_ai ecosystem.
"""

import os
from typing import Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from flo_ai.llm import BaseLLM


class LLMFactory:
    """Factory class for creating LLM instances from configuration."""

    SUPPORTED_PROVIDERS = {
        'openai',
        'anthropic',
        'gemini',
        'ollama',
        'vertexai',
        'rootflo',
    }

    @staticmethod
    def create_llm(model_config: Dict[str, Any], **kwargs) -> 'BaseLLM':
        """Create an LLM instance from model configuration.

        Args:
            model_config: Dictionary containing model configuration with keys:
                - provider (str): LLM provider name (default: 'openai')
                - name (str): Model name (required for most providers)
                - base_url (str, optional): Custom base URL
                - model_id (str): For RootFlo provider
                - project (str): For VertexAI provider
                - location (str): For VertexAI provider (default: 'asia-south1')
            **kwargs: Additional parameters that override config and env vars:
                - base_url: Override base URL
                - For RootFlo: app_key, app_secret, issuer, audience, access_token

        Returns:
            BaseLLM: Configured LLM instance

        Raises:
            ValueError: If provider is unsupported or required parameters are missing

        Examples:
            >>> # OpenAI
            >>> llm = LLMFactory.create_llm({'provider': 'openai', 'name': 'gpt-4'})

            >>> # VertexAI with project
            >>> llm = LLMFactory.create_llm({
            ...     'provider': 'vertexai',
            ...     'name': 'gemini-pro',
            ...     'project': 'my-project',
            ...     'location': 'us-central1'
            ... })

            >>> # RootFlo with auth
            >>> llm = LLMFactory.create_llm(
            ...     {'provider': 'rootflo', 'model_id': 'model-123'},
            ...     app_key='key', app_secret='secret', issuer='iss', audience='aud'
            ... )
        """
        provider = model_config.get('provider', 'openai').lower()

        if provider not in LLMFactory.SUPPORTED_PROVIDERS:
            raise ValueError(
                f'Unsupported model provider: {provider}. '
                f'Supported providers: {", ".join(sorted(LLMFactory.SUPPORTED_PROVIDERS))}'
            )

        if provider == 'rootflo':
            return LLMFactory._create_rootflo_llm(model_config, **kwargs)
        elif provider == 'vertexai':
            return LLMFactory._create_vertexai_llm(model_config, **kwargs)
        else:
            return LLMFactory._create_standard_llm(provider, model_config, **kwargs)

    @staticmethod
    def _create_standard_llm(
        provider: str, model_config: Dict[str, Any], **kwargs
    ) -> 'BaseLLM':
        """Create standard LLM instances (OpenAI, Anthropic, Gemini, Ollama)."""
        from flo_ai.llm import OpenAI, Anthropic, Gemini, OllamaLLM

        model_name = model_config.get('name')
        if not model_name:
            raise ValueError(
                f'{provider.title()} provider requires "name" parameter in model configuration'
            )

        # Priority: kwargs > model_config > None
        base_url = kwargs.get('base_url') or model_config.get('base_url')

        provider_map = {
            'openai': OpenAI,
            'anthropic': Anthropic,
            'gemini': Gemini,
            'ollama': OllamaLLM,
        }

        llm_class = provider_map[provider]
        return llm_class(model=model_name, base_url=base_url)

    @staticmethod
    def _create_vertexai_llm(model_config: Dict[str, Any], **kwargs) -> 'BaseLLM':
        """Create VertexAI LLM instance with project and location."""
        from flo_ai.llm import VertexAI

        model_name = model_config.get('name')
        if not model_name:
            raise ValueError(
                'VertexAI provider requires "name" parameter in model configuration'
            )

        # Get VertexAI-specific parameters
        project = kwargs.get('project') or model_config.get('project')
        location = kwargs.get('location') or model_config.get('location', 'asia-south1')
        base_url = kwargs.get('base_url') or model_config.get('base_url')

        if not project:
            raise ValueError(
                'VertexAI provider requires "project" parameter. '
                'Provide it in model_config or as a kwarg.'
            )

        return VertexAI(
            model=model_name,
            project=project,
            location=location,
            base_url=base_url,
        )

    @staticmethod
    def _create_rootflo_llm(model_config: Dict[str, Any], **kwargs) -> 'BaseLLM':
        """Create RootFlo LLM instance with authentication."""
        from flo_ai.llm import RootFloLLM

        model_id = model_config.get('model_id')
        if not model_id:
            raise ValueError(
                'RootFlo provider requires "model_id" in model configuration'
            )

        # Gather RootFlo parameters from kwargs or environment
        base_url = (
            kwargs.get('base_url')
            or model_config.get('base_url')
            or os.getenv('ROOTFLO_BASE_URL')
        )
        app_key = kwargs.get('app_key') or os.getenv('ROOTFLO_APP_KEY')
        app_secret = kwargs.get('app_secret') or os.getenv('ROOTFLO_APP_SECRET')
        issuer = kwargs.get('issuer') or os.getenv('ROOTFLO_ISSUER')
        audience = kwargs.get('audience') or os.getenv('ROOTFLO_AUDIENCE')
        access_token = kwargs.get('access_token')  # Optional, from kwargs only

        # Validate required parameters based on auth method
        if not access_token:
            # JWT auth flow - requires all parameters
            required_params = {
                'base_url': base_url,
                'app_key': app_key,
                'app_secret': app_secret,
                'issuer': issuer,
                'audience': audience,
            }
            missing = [k for k, v in required_params.items() if not v]

            if missing:
                raise ValueError(
                    f'RootFlo configuration incomplete. Missing required parameters: {", ".join(missing)}. '
                    f'These can be provided via kwargs or environment variables '
                    f'(ROOTFLO_BASE_URL, ROOTFLO_APP_KEY, ROOTFLO_APP_SECRET, ROOTFLO_ISSUER, ROOTFLO_AUDIENCE).'
                )
        else:
            # Access token flow - only needs base_url and app_key
            required_params = {
                'base_url': base_url,
                'app_key': app_key,
            }
            missing = [k for k, v in required_params.items() if not v]

            if missing:
                raise ValueError(
                    f'RootFlo configuration incomplete. Missing required parameters: {", ".join(missing)}. '
                    f'These can be provided via kwargs or environment variables '
                    f'(ROOTFLO_BASE_URL, ROOTFLO_APP_KEY).'
                )

        return RootFloLLM(
            base_url=base_url,
            model_id=model_id,
            app_key=app_key,
            app_secret=app_secret,
            issuer=issuer,
            audience=audience,
            access_token=access_token,
        )


# Convenience function for direct import
def create_llm_from_config(model_config: Dict[str, Any], **kwargs) -> 'BaseLLM':
    """
    Convenience function to create an LLM instance from configuration.

    This is a wrapper around LLMFactory.create_llm() for easier imports.

    Args:
        model_config: Dictionary containing model configuration
        **kwargs: Additional parameters that override config and env vars

    Returns:
        BaseLLM: Configured LLM instance

    See LLMFactory.create_llm() for detailed documentation.
    """
    return LLMFactory.create_llm(model_config, **kwargs)



from enum import Enum
from typing import AsyncIterator, Dict, Any, List, Optional
from datetime import datetime, timedelta
from flo_ai.models.chat_message import ImageMessageContent
import jwt
import httpx
from .base_llm import BaseLLM
from .openai_llm import OpenAI
from .gemini_llm import Gemini
from .anthropic_llm import Anthropic
from flo_ai.tool.base_tool import Tool


class LLMProvider(Enum):
    """Enum for supported LLM providers"""

    OPENAI = 'openai'
    GEMINI = 'gemini'
    ANTHROPIC = 'anthropic'


class RootFloLLM(BaseLLM):
    """
    Proxy LLM class that routes to different SDK implementations based on type.
    Acts as a unified interface to OpenAI, Gemini, and Anthropic SDKs via a proxy URL.
    """

    def __init__(
        self,
        base_url: str,
        model_id: str,
        app_key: str,
        app_secret: str,
        issuer: str,
        audience: str,
        access_token: Optional[str] = None,
        temperature: float = 0.7,
        **kwargs,
    ):
        """
        Initialize RootFloLLM proxy.

        Args:
            base_url: The base URL of the proxy server
            model_id: The model identifier (config_id)
            app_key: Application key for X-Rootflo-Key header
            app_secret: Application secret for JWT signing
            issuer: JWT issuer claim
            audience: JWT audience claim
            access_token: Optional pre-generated access token (if provided, skips JWT generation)
            temperature: Temperature parameter for generation
            **kwargs: Additional parameters to pass to the underlying SDK
        """
        # Validate required parameters

        if not model_id:
            raise ValueError('model_id is required')

        if not base_url:
            raise ValueError('base_url is required')

        # Validate JWT credentials if access_token is not provided
        if not access_token:
            missing = []
            if not app_key:
                missing.append('app_key')
            if not app_secret:
                missing.append('app_secret')
            if not issuer:
                missing.append('issuer')
            if not audience:
                missing.append('audience')

            if missing:
                raise ValueError(
                    f'Missing required parameters for JWT generation: {", ".join(missing)}. '
                    f'Either provide these parameters or pass an access_token directly.'
                )
        else:  # app key is still required
            if not app_key:
                raise ValueError('app_key is required even when using access_token')

        # Use provided access_token or generate JWT token
        if access_token:
            api_token = access_token
        else:
            now = datetime.now()
            payload = {
                'iss': issuer,
                'aud': audience,
                'iat': int(now.timestamp()),
                'exp': int((now + timedelta(seconds=3600)).timestamp()),
                'role_id': 'floconsole-service',
                'user_id': 'service',
                'service_auth': True,
            }
            service_token = jwt.encode(payload, app_secret, algorithm='HS256')
            api_token = f'fc_{service_token}'

        # Fetch LLM configuration from API
        config = self._fetch_llm_config(base_url, model_id, api_token, app_key)
        llm_model = config['llm_model']
        llm_type = config['type']

        # Map type string to LLMProvider enum
        try:
            llm_provider = LLMProvider(llm_type.lower())
        except ValueError:
            raise ValueError(
                f'Unsupported LLM provider type from API: {llm_type}. '
                f'Supported types: {[p.value for p in LLMProvider]}'
            )

        super().__init__(
            model=llm_model, api_key=api_token, temperature=temperature, **kwargs
        )

        self.base_url = base_url
        self.model_id = model_id
        self.llm_provider = llm_provider

        # Construct full URL for LLM inference
        full_url = f'{base_url}/v1/llm-inference/{model_id}'

        # Prepare custom headers for proxy authentication
        custom_headers = {'X-Rootflo-Key': app_key}

        # Instantiate appropriate SDK wrapper based on llm_provider
        if llm_provider == LLMProvider.OPENAI:
            self._llm = OpenAI(
                model=llm_model,
                base_url=full_url,
                api_key=api_token,
                temperature=temperature,
                custom_headers=custom_headers,
                **kwargs,
            )
        elif llm_provider == LLMProvider.ANTHROPIC:
            self._llm = Anthropic(
                model=llm_model,
                base_url=full_url,
                api_key=api_token,
                temperature=temperature,
                custom_headers=custom_headers,
                **kwargs,
            )
        elif llm_provider == LLMProvider.GEMINI:
            # Gemini SDK - pass base_url which will be handled via http_options
            self._llm = Gemini(
                model=llm_model,
                api_key=api_token,
                temperature=temperature,
                base_url=full_url,
                custom_headers=custom_headers,
                **kwargs,
            )
        else:
            raise ValueError(f'Unsupported LLM provider: {llm_provider}')

    def _fetch_llm_config(
        self, base_url: str, model_id: str, api_token: str, app_key: str
    ) -> Dict[str, Any]:
        """
        Fetch LLM configuration from the API endpoint.

        Args:
            base_url: The base URL of the API server
            model_id: The model identifier (config_id)
            api_token: The JWT token for authorization
            app_key: Application key for X-Rootflo-Key header

        Returns:
            Dict containing llm_model and type

        Raises:
            Exception: If API call fails or response is invalid
        """
        config_url = f'{base_url}/v1/llm-inference-configs/{model_id}'
        headers = {
            'Authorization': f'Bearer {api_token}',
            'X-Rootflo-Key': app_key,
        }

        try:
            with httpx.Client() as client:
                response = client.get(config_url, headers=headers, timeout=30.0)
                response.raise_for_status()

                data = response.json()

                config_data = data.get('data')
                if not config_data:
                    raise Exception('API response missing data field')

                llm_model = config_data.get('llm_model')
                llm_type = config_data.get('type')

                if not llm_model or not llm_type:
                    raise Exception(
                        f'API response missing required fields: llm_model={llm_model}, type={llm_type}'
                    )

                return {'llm_model': llm_model, 'type': llm_type}

        except httpx.HTTPStatusError as e:
            raise Exception(
                f'API request failed with status {e.response.status_code}: {e.response.text}'
            ) from e
        except httpx.RequestError as e:
            raise Exception(f'API request failed: {str(e)}') from e
        except Exception as e:
            raise Exception(f'Failed to fetch LLM config: {str(e)}') from e

    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Generate a response from the LLM"""
        return await self._llm.generate(
            messages, functions=functions, output_schema=output_schema, **kwargs
        )

    async def stream(
        self,
        messages: List[Dict[str, Any]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Generate a streaming response from the LLM"""
        async for chunk in self._llm.stream(messages, functions=functions, **kwargs):
            yield chunk

    def get_message_content(self, response: Any) -> str:
        """Extract message content from response"""
        return self._llm.get_message_content(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a tool for the specific LLM's API"""
        return self._llm.format_tool_for_llm(tool)

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format a list of tools for the specific LLM's API"""
        return self._llm.format_tools_for_llm(tools)

    def format_image_in_message(self, image: ImageMessageContent) -> str:
        """Format a image in the message"""
        return self._llm.format_image_in_message(image)



from typing import Literal, Optional, Dict, Any
from dataclasses import dataclass, field


class MessageType:
    USER = 'user'
    ASSISTANT = 'assistant'
    FUNCTION = 'function'
    SYSTEM = 'system'


@dataclass
class MediaMessageContent:
    type: Optional[Literal['text', 'image', 'document']] = None
    url: Optional[str] = None
    base64: Optional[str] = None
    mime_type: Optional[str] = None


@dataclass
class ImageMessageContent(MediaMessageContent):
    url: Optional[str] = None
    base64: Optional[str] = None
    mime_type: Optional[str] = None
    bytes: Optional[bytes] = None

    def __post_init__(self):
        self.type = 'image'


@dataclass
class DocumentMessageContent(MediaMessageContent):
    url: Optional[str] = None
    base64: Optional[str] = None
    bytes: Optional[bytes] = None
    mime_type: Optional[str] = None

    def __post_init__(self):
        self.type = 'document'


@dataclass
class TextMessageContent:
    text: str
    type: str = 'text'


@dataclass
class BaseMessage:
    content: str | ImageMessageContent | DocumentMessageContent | TextMessageContent
    role: Optional[Literal['system', 'user', 'assistant', 'function']] = None
    metadata: Optional[Dict[str, Any]] = None


@dataclass
class SystemMessage(BaseMessage):
    content: str
    metadata: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        self.role = 'system'


@dataclass
class UserMessage(BaseMessage):
    content: str | ImageMessageContent | DocumentMessageContent | TextMessageContent
    metadata: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        self.role = 'user'


@dataclass
class AssistantMessage(BaseMessage):
    content: str
    metadata: Optional[Dict[str, Any]] = None
    role: Optional[str] = None

    def __post_init__(self):
        if self.role is None:
            self.role = MessageType.ASSISTANT


@dataclass
class FunctionMessage(BaseMessage):
    """Message representing a function/tool call result.

    According to OpenAI's API, function results should use:
    {
        "role": "function",
        "name": "<function-name>",
        "content": "<result>"
    }
    """

    content: str  # Override parent's content (must come before fields with defaults)
    name: str = field(
        kw_only=True
    )  # Function/tool name that was called (keyword-only to allow it after defaults)

    def __post_init__(self):
        self.role = MessageType.FUNCTION



from typing import Dict, Any, List, Optional, Tuple
from abc import ABC, abstractmethod
from enum import Enum
from flo_ai.llm.base_llm import BaseLLM
from flo_ai.models.chat_message import (
    BaseMessage,
    MediaMessageContent,
    TextMessageContent,
    FunctionMessage,
)
from flo_ai.utils.variable_extractor import resolve_variables


class AgentType(Enum):
    CONVERSATIONAL = 'conversational'
    TOOL_USING = 'tool_using'


class ReasoningPattern(Enum):
    DIRECT = 'direct'  # Direct response without explicit reasoning
    REACT = 'react'  # Thought-Action-Observation cycle
    COT = 'cot'  # Chain of Thought reasoning


class BaseAgent(ABC):
    def __init__(
        self,
        name: str,
        system_prompt: str,
        agent_type: AgentType,
        llm: BaseLLM,
        max_retries: int = 3,
        max_tool_calls: int = 5,
    ):
        self.name = name
        self.system_prompt = system_prompt
        self.agent_type = agent_type
        self.llm = llm
        self.max_retries = max_retries
        self.max_tool_calls = max_tool_calls
        self.resolved_variables = False
        self.conversation_history: List[BaseMessage] = []

    @abstractmethod
    async def run(self, input_text: str) -> str:
        """Execute the agent's main functionality"""
        pass

    async def handle_error(
        self, error: Exception, context: Dict[str, Any]
    ) -> Tuple[bool, str]:
        error_prompt = (
            f'An error occurred while processing the request: {str(error)}\n'
            f'Context: {context}\n'
            'Please analyze the error and suggest a correction. '
            'If the error is not recoverable, explain why.'
        )

        try:
            messages = [
                {
                    'role': 'system',
                    'content': 'You are an AI error analysis assistant. '
                    'Analyze errors and suggest corrections when possible.',
                },
                {'role': 'user', 'content': error_prompt},
            ]

            response = await self.llm.generate(messages)
            analysis = self.llm.get_message_content(response)
            should_retry = 'not recoverable' not in analysis.lower()
            return should_retry, analysis

        except Exception as e:
            return False, f'Error during error handling: {str(e)}'

    def add_to_history(self, input_message: BaseMessage | List[BaseMessage]):
        if isinstance(input_message, list):
            self.conversation_history.extend(input_message)
        else:
            self.conversation_history.append(input_message)

    def clear_history(self):
        """Clear conversation history"""
        self.conversation_history = []

    async def _get_message_history(self, variables: Optional[Dict[str, Any]] = None):
        message_history = []
        for input in self.conversation_history:
            # Handle FunctionMessage (OpenAI function role format)
            if isinstance(input, FunctionMessage):
                message_history.append(
                    {'role': input.role, 'name': input.name, 'content': input.content}
                )
            # CRITICAL: Check content type FIRST, before message type
            # This ensures TextMessageContent objects are converted to strings
            elif isinstance(input.content, TextMessageContent):
                resolved_content = resolve_variables(input.content.text, variables)
                message_history.append(
                    {'role': input.role, 'content': resolved_content}
                )
            elif isinstance(input.content, MediaMessageContent):
                if input.content.type == 'image':
                    # Format image message and add to history
                    formatted_content = self.llm.format_image_in_message(input.content)
                    message_history.append(
                        {'role': input.role, 'content': formatted_content}
                    )

                elif input.content.type == 'document':
                    # Format document message and add to history
                    formatted_content = await self.llm.format_document_in_message(
                        input.content
                    )
                    message_history.append(
                        {'role': input.role, 'content': formatted_content}
                    )
                else:
                    raise ValueError(
                        f'Invalid media message content type: {input.content.type}'
                    )
            elif isinstance(input.content, str):
                # Handle other messages with string content (UserMessage, SystemMessage, etc.)
                resolved_content = resolve_variables(input.content, variables)
                message_history.append(
                    {'role': input.role, 'content': resolved_content}
                )
            else:
                raise ValueError(f'Invalid content type: {type(input.content)}')
        return message_history



from abc import ABC, abstractmethod
from typing import Dict, Any, List, Optional, AsyncIterator
from flo_ai.tool.base_tool import Tool
from flo_ai.utils.document_processor import get_default_processor
from flo_ai.utils.logger import logger
from flo_ai.models.chat_message import DocumentMessageContent, ImageMessageContent


class BaseLLM(ABC):
    def __init__(
        self, model: str, api_key: str = None, temperature: float = 0.7, **kwargs
    ):
        self.model = model
        self.api_key = api_key
        self.temperature = temperature
        self.kwargs = kwargs

    @abstractmethod
    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Generate a response from the LLM"""
        pass

    @abstractmethod
    async def stream(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from the LLM as they are generated"""
        pass

    async def get_function_call(
        self, response: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        """Extract function call information from LLM response"""
        if hasattr(response, 'function_call') and response.function_call:
            result = {
                'name': response.function_call.name,
                'arguments': response.function_call.arguments,
            }
            # Include ID if available (LLM-specific)
            if hasattr(response.function_call, 'id'):
                result['id'] = response.function_call.id
            return result
        elif isinstance(response, dict) and 'function_call' in response:
            result = {
                'name': response['function_call']['name'],
                'arguments': response['function_call']['arguments'],
            }
            # Include ID if available (LLM-specific)
            if 'id' in response['function_call']:
                result['id'] = response['function_call']['id']
            return result
        return None

    def get_assistant_message_for_tool_call(
        self, response: Dict[str, Any]
    ) -> Optional[Any]:
        """
        Get the assistant message content for tool calls.
        Override in LLM-specific implementations if special handling is needed.
        Returns None to use default text content extraction.
        """
        return None

    def get_tool_use_id(self, function_call: Dict[str, Any]) -> Optional[str]:
        """
        Extract tool_use_id from function call if available.
        Override in LLM-specific implementations if IDs are used.
        Returns None by default.
        """
        return function_call.get('id')

    def format_function_result_message(
        self, function_name: str, content: str, tool_use_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Format a function result message for the LLM.
        Override in LLM-specific implementations for special formatting.
        """
        message = {
            'role': 'function',
            'name': function_name,
            'content': content,
        }
        if tool_use_id:
            message['tool_use_id'] = tool_use_id
        return message

    @abstractmethod
    def get_message_content(self, response: Dict[str, Any]) -> str:
        """Extract message content from response"""
        pass

    @abstractmethod
    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a tool for the specific LLM's API"""
        pass

    @abstractmethod
    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format a list of tools for the specific LLM's API"""
        pass

    @abstractmethod
    def format_image_in_message(self, image: ImageMessageContent) -> str:
        """Format a image in the message"""
        pass

    async def format_document_in_message(self, document: DocumentMessageContent) -> str:
        """Format a document in the message by extracting text content"""

        try:
            # Process document to extract text
            result = await get_default_processor().process_document(document)

            # Format the extracted content for the LLM
            extracted_text = result.get('extracted_text', '')
            doc_type = result.get('document_type', 'unknown')

            logger.info(
                f'Successfully formatted {doc_type} document for {self.__class__.__name__} LLM'
            )
            return extracted_text

        except Exception as e:
            logger.error(
                f'Error formatting document for {self.__class__.__name__}: {e}'
            )
            raise Exception(f'Failed to format document: {str(e)}')



"""
Plan Execution Tools for Flo AI Framework

This module provides reusable tools for plan-based execution patterns,
enabling agents to create, store, and manage execution plans automatically.
"""

import uuid
import re
from flo_ai.tool.base_tool import Tool
from flo_ai.arium.memory import PlanAwareMemory, ExecutionPlan, PlanStep, StepStatus


class PlanTool(Tool):
    """Tool for creating and storing execution plans in PlanAwareMemory"""

    def __init__(self, memory: PlanAwareMemory):
        """
        Initialize the plan tool.

        Args:
            memory: PlanAwareMemory instance to store plans in
        """
        self.memory = memory
        super().__init__(
            name='store_execution_plan',
            description='Create and store an execution plan from plan text. Use this after generating a plan.',
            function=self._execute_plan_storage,
            parameters={
                'plan_text': {
                    'type': 'string',
                    'description': 'The generated plan text in the required format',
                }
            },
        )

    async def _execute_plan_storage(self, plan_text: str) -> str:
        """Parse plan text and store ExecutionPlan object in memory"""
        try:
            execution_plan = self._parse_plan_text(plan_text)

            if execution_plan:
                self.memory.add_plan(execution_plan)

                plan_summary = f'✅ Plan stored: {execution_plan.title}\n'
                plan_summary += f'📊 Steps: {len(execution_plan.steps)}\n'

                for i, step in enumerate(execution_plan.steps, 1):
                    deps = (
                        f" (depends: {', '.join(step.dependencies)})"
                        if step.dependencies
                        else ''
                    )
                    plan_summary += (
                        f'  {i}. {step.id}: {step.description} → {step.agent}{deps}\n'
                    )

                return plan_summary
            else:
                return '❌ Failed to parse plan text into ExecutionPlan'

        except Exception as e:
            return f'❌ Error storing plan: {str(e)}'

    def _parse_plan_text(self, plan_text: str) -> ExecutionPlan:
        """Parse LLM-generated plan text into ExecutionPlan object"""

        # Extract title
        title_match = re.search(r'EXECUTION PLAN:\s*(.+)', plan_text)
        title = title_match.group(1).strip() if title_match else 'Generated Plan'

        # Extract description
        desc_match = re.search(r'DESCRIPTION:\s*(.+)', plan_text)
        description = desc_match.group(1).strip() if desc_match else 'Execution plan'

        # Extract steps using regex
        steps = []
        step_pattern = (
            r'(\d+)\.\s*(\w+):\s*(.+?)\s*→\s*(\w+)(?:\s*\(depends on:\s*([^)]+)\))?'
        )

        for match in re.finditer(step_pattern, plan_text, re.MULTILINE):
            step_num, step_id, step_desc, agent, deps_str = match.groups()

            dependencies = []
            if deps_str:
                dependencies = [dep.strip() for dep in deps_str.split(',')]

            step = PlanStep(
                id=step_id,
                description=step_desc.strip(),
                agent=agent,
                dependencies=dependencies,
                status=StepStatus.PENDING,
            )
            steps.append(step)

        return ExecutionPlan(
            id=str(uuid.uuid4()),
            title=title,
            description=description,
            steps=steps,
            created_by='planner',
        )


class StepTool(Tool):
    """Tool for marking execution steps as completed"""

    def __init__(self, memory: PlanAwareMemory, agent_name: str):
        """
        Initialize the step tool.

        Args:
            memory: PlanAwareMemory instance to update plans in
            agent_name: Name of the agent this tool belongs to
        """
        self.memory = memory
        self.agent_name = agent_name
        super().__init__(
            name='complete_step',
            description='Mark a plan step as completed after executing it',
            function=self._execute_step_completion,
            parameters={
                'step_id': {
                    'type': 'string',
                    'description': 'The ID of the step that was completed',
                },
                'result': {
                    'type': 'string',
                    'description': 'The result or output of completing the step',
                },
            },
        )

    async def _execute_step_completion(self, step_id: str, result: str) -> str:
        """Mark a step as completed and store the result"""
        try:
            current_plan = self.memory.get_current_plan()
            if not current_plan:
                return '❌ No current plan found'

            step = current_plan.get_step(step_id)
            if not step:
                return f'❌ Step {step_id} not found in current plan'

            if step.agent != self.agent_name:
                return f'❌ Step {step_id} is assigned to {step.agent}, not {self.agent_name}'

            # Mark step as completed
            step.status = StepStatus.COMPLETED
            step.result = result
            self.memory.update_plan(current_plan)

            return f'✅ Step {step_id} marked as completed'

        except Exception as e:
            return f'❌ Error completing step: {str(e)}'


class PlanStatusTool(Tool):
    """Tool for checking the current plan status and next steps"""

    def __init__(self, memory: PlanAwareMemory):
        """
        Initialize the plan status tool.

        Args:
            memory: PlanAwareMemory instance to check
        """
        self.memory = memory
        super().__init__(
            name='check_plan_status',
            description='Check the current execution plan status and get next available steps',
            function=self._execute_status_check,
            parameters={},
        )

    async def _execute_status_check(self) -> str:
        """Get current plan status and next steps"""
        try:
            current_plan = self.memory.get_current_plan()
            if not current_plan:
                return '❌ No current execution plan found'

            # Get plan overview
            status_info = f'📋 Current Plan: {current_plan.title}\n'
            status_info += f'📝 Description: {current_plan.description}\n'
            status_info += f'✅ Completed: {current_plan.is_completed()}\n'

            # Get next steps
            next_steps = current_plan.get_next_steps()
            if next_steps:
                status_info += f'\n🎯 Next Steps ({len(next_steps)} available):\n'
                for step in next_steps:
                    deps = (
                        f" (depends: {', '.join(step.dependencies)})"
                        if step.dependencies
                        else ''
                    )
                    status_info += (
                        f'  • {step.id}: {step.description} → {step.agent}{deps}\n'
                    )
            else:
                if current_plan.is_completed():
                    status_info += '\n🎉 All steps completed!'
                else:
                    status_info += '\n⏳ Waiting for dependencies to complete'

            # Show all steps with status
            status_info += '\n📊 All Steps:\n'
            for step in current_plan.steps:
                status_icon = {
                    StepStatus.PENDING: '○',
                    StepStatus.IN_PROGRESS: '⏳',
                    StepStatus.COMPLETED: '✅',
                    StepStatus.FAILED: '❌',
                }.get(step.status, '○')
                status_info += (
                    f'  {status_icon} {step.id}: {step.description} → {step.agent}\n'
                )

            return status_info

        except Exception as e:
            return f'❌ Error checking plan status: {str(e)}'



from typing import Protocol, runtime_checkable, List, Any, Dict, Optional


@runtime_checkable
class ExecutableNode(Protocol):
    """
    Protocol defining the interface for any node that can be executed
    within an Arium workflow.

    Any class implementing this protocol can be used as a node:
    - Agent (already implements)
    - Tool (already implements)
    - Arium (already implements!)
    - Custom node types
    """

    name: str
    """Unique identifier for the node"""
    input_filter: Optional[List[str]] = None
    """List of input keys to include in the node's execution"""

    async def run(
        self,
        inputs: List[str | Any],
        variables: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Any:
        """
        Execute the node and return results.

        Args:
            inputs: Input data for execution
            variables: Optional variable substitutions
            **kwargs: Additional execution parameters

        Returns:
            Execution result (type depends on node implementation)
        """
        ...



"""
OpenTelemetry integration for flo_ai
"""

from .telemetry import (
    FloTelemetry,
    get_tracer,
    get_meter,
    configure_telemetry,
    shutdown_telemetry,
)

__all__ = [
    'FloTelemetry',
    'get_tracer',
    'get_meter',
    'configure_telemetry',
    'shutdown_telemetry',
]



"""
Document-related data models for Flo AI framework.

This module contains document types and message classes to avoid circular imports.
"""

from enum import Enum


class DocumentType(Enum):
    """Enumeration of supported document types."""

    PDF = 'application/pdf'
    TXT = 'text/plain'



"""
Models package for flo_ai - Agent framework components
"""

from .agent import Agent, MessageType
from .agent_error import AgentError
from .base_agent import BaseAgent, AgentType, ReasoningPattern
from .document import DocumentType
from .chat_message import (
    SystemMessage,
    UserMessage,
    AssistantMessage,
    FunctionMessage,
    BaseMessage,
    MediaMessageContent,
    TextMessageContent,
    ImageMessageContent,
    DocumentMessageContent,
)

__all__ = [
    'Agent',
    'AgentError',
    'BaseAgent',
    'AgentType',
    'ReasoningPattern',
    'DocumentType',
    'MessageType',
    'SystemMessage',
    'UserMessage',
    'AssistantMessage',
    'FunctionMessage',
    'BaseMessage',
    'MediaMessageContent',
    'TextMessageContent',
    'ImageMessageContent',
    'DocumentMessageContent',
]



from dataclasses import dataclass
from typing import Callable, List, Optional
from functools import partial
from flo_ai.arium.memory import BaseMemory


def default_router(
    to_node: str,
    memory: BaseMemory,
) -> str:
    return to_node


@dataclass
class StartNode:
    name = '__start__'
    input_filter: Optional[List[str]] = None


@dataclass
class EndNode:
    name = '__end__'
    input_filter: Optional[List[str]] = None


@dataclass
class Edge:
    router_fn: Callable | partial
    to_nodes: List[str]

    def is_default_router(self) -> bool:
        if isinstance(self.router_fn, partial):
            return self.router_fn.func.__name__ == 'default_router'
        return False



"""
Document processing utilities for Flo AI framework.

This module provides extensible document processing capabilities for PDF and TXT files,
with a factory pattern design for easy addition of new document types.
"""

import base64
import time
from abc import ABC, abstractmethod
from typing import Dict, Any, Union

import pymupdf
import pymupdf4llm
import chardet

from flo_ai.models.document import DocumentType
from flo_ai.models.chat_message import DocumentMessageContent
from flo_ai.utils.logger import logger


class DocumentProcessingError(Exception):
    """Exception raised when document processing fails."""

    pass


class BaseDocumentProcessor(ABC):
    """Abstract base class for document processors."""

    @abstractmethod
    async def process(self, document: DocumentMessageContent) -> Dict[str, Any]:
        """
        Process a document and return extracted content and metadata.

        Args:
            document: DocumentMessageContent containing document data

        Returns:
            Dict containing extracted text, metadata, and processing info
        """
        pass


class PDFProcessor(BaseDocumentProcessor):
    """Processor for PDF documents."""

    async def process(self, document: DocumentMessageContent) -> Dict[str, Any]:
        """Extract text and metadata from PDF document."""
        try:
            pdf_content = await self._get_pdf_content(document)

            # Process with pymupdf4llm (LLM-optimized)
            text_data = await self._process_with_pymupdf4llm(pdf_content)

            return {
                'extracted_text': text_data['text'],
                'page_count': text_data.get('page_count', 0),
                'processing_method': text_data.get('method', 'unknown'),
                'metadata': text_data.get('metadata', {}),
                'document_type': DocumentType.PDF.value,
            }

        except Exception as e:
            logger.error(f'Error processing PDF: {str(e)}')
            raise DocumentProcessingError(f'Failed to process PDF: {str(e)}')

    async def _get_pdf_content(
        self, document: DocumentMessageContent
    ) -> Union[str, bytes]:
        """Get PDF content from various sources."""
        if document.bytes:
            return document.bytes
        elif document.base64:
            return base64.b64decode(document.base64)
        elif document.url:
            return document.url
        else:
            raise DocumentProcessingError('No PDF content provided')

    async def _process_with_pymupdf4llm(
        self, pdf_content: Union[str, bytes]
    ) -> Dict[str, Any]:
        """Process PDF using pymupdf4llm (LLM-optimized)."""
        if isinstance(pdf_content, str):
            # File path - pass directly to pymupdf4llm
            text_data = pymupdf4llm.to_markdown(pdf_content)
            metadata = {}
        else:
            # Bytes - create PyMuPDF Document from memory
            doc = pymupdf.open(stream=pdf_content)
            try:
                text_data = pymupdf4llm.to_markdown(doc)
                metadata = {}
            finally:
                doc.close()  # Clean up document object

        return {
            'text': text_data,
            'method': 'pymupdf4llm',
            'metadata': metadata,
            'page_count': len(text_data.split('\n---\n')) if '---' in text_data else 1,
        }


class TXTProcessor(BaseDocumentProcessor):
    """Processor for text documents."""

    async def process(self, document: DocumentMessageContent) -> Dict[str, Any]:
        """Extract text from TXT document."""
        try:
            text_content = await self._get_text_content(document)

            return {
                'extracted_text': text_content,
                'page_count': 1,
                'processing_method': 'text_reader',
                'metadata': {
                    'character_count': len(text_content),
                    'line_count': len(text_content.splitlines()),
                    'encoding': 'utf-8',
                },
                'document_type': DocumentType.TXT.value,
            }

        except Exception as e:
            logger.error(f'Error processing TXT: {str(e)}')
            raise DocumentProcessingError(f'Failed to process TXT: {str(e)}')

    async def _get_text_content(self, document: DocumentMessageContent) -> str:
        """Get text content from various sources."""
        if document.bytes:
            return await self._decode_bytes(document.bytes)
        elif document.base64:
            decoded_bytes = base64.b64decode(document.base64)
            return await self._decode_bytes(decoded_bytes)
        else:
            raise DocumentProcessingError('No TXT content provided')

    async def _read_text_file(self, file_path: str) -> str:
        """Read text file with encoding detection."""
        try:
            # Try UTF-8 first
            with open(file_path, 'r', encoding='utf-8') as f:
                return f.read()
        except UnicodeDecodeError:
            # Try encoding detection with chardet
            with open(file_path, 'rb') as f:
                raw_data = f.read()
            detected = chardet.detect(raw_data)
            encoding = detected.get('encoding', 'utf-8')
            return raw_data.decode(encoding, errors='replace')

    async def _decode_bytes(self, content_bytes: bytes) -> str:
        """Decode bytes with encoding detection."""
        try:
            return content_bytes.decode('utf-8')
        except UnicodeDecodeError:
            detected = chardet.detect(content_bytes)
            encoding = detected.get('encoding', 'utf-8')
            return content_bytes.decode(encoding, errors='replace')


class DocumentProcessor:
    """
    Main document processor with factory pattern for extensibility.

    Supports PDF and TXT documents with easy extension for new types.
    """

    def __init__(self):
        self._processors = {
            DocumentType.PDF: PDFProcessor(),
            DocumentType.TXT: TXTProcessor(),
        }

    def register_processor(
        self, document_type: DocumentType, processor: BaseDocumentProcessor
    ):
        """Register a new document processor for a specific type."""
        self._processors[document_type] = processor

    async def process_document(
        self, document: DocumentMessageContent
    ) -> Dict[str, Any]:
        """
        Process a document using the appropriate processor.

        Args:
            document: DocumentMessageContent containing document data

        Returns:
            Dict containing extracted content and metadata

        Raises:
            DocumentProcessingError: If processing fails or document type unsupported
        """
        # Convert mime_type string to DocumentType enum
        if not document.mime_type:
            raise DocumentProcessingError('Document mime_type is required')

        # Map mime_type string to DocumentType enum
        document_type = None
        for doc_type in DocumentType:
            if doc_type.value == document.mime_type:
                document_type = doc_type
                break

        if document_type is None or document_type not in self._processors:
            raise DocumentProcessingError(
                f'Unsupported document type: {document.mime_type}. '
                f'Supported types: {[dt.value for dt in self._processors.keys()]}'
            )

        processor: BaseDocumentProcessor = self._processors[document_type]

        try:
            result = await processor.process(document)

            # Add common metadata
            result['processing_timestamp'] = time.time()

            logger.info(
                f"Successfully processed {document_type.value} document "
                f"using {result.get('processing_method', 'unknown')} method"
            )

            return result

        except Exception as e:
            logger.error(f'Document processing failed: {str(e)}')
            raise


# Lazy singleton for default processor
_default_processor = None


def get_default_processor() -> DocumentProcessor:
    """Get the default DocumentProcessor instance (lazy singleton)."""
    global _default_processor
    if _default_processor is None:
        _default_processor = DocumentProcessor()
    return _default_processor



"""
OpenTelemetry telemetry implementation for flo_ai framework
"""

from typing import Optional, Dict, Any
from opentelemetry import trace, metrics
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import (
    PeriodicExportingMetricReader,
    ConsoleMetricExporter,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
import os


class FloTelemetry:
    """
    Central telemetry configuration for flo_ai framework.

    Provides OpenTelemetry integration with tracing and metrics support.
    """

    _instance: Optional['FloTelemetry'] = None
    _initialized: bool = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(FloTelemetry, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not FloTelemetry._initialized:
            self.tracer_provider: Optional[TracerProvider] = None
            self.meter_provider: Optional[MeterProvider] = None
            self.tracer: Optional[trace.Tracer] = None
            self.meter: Optional[metrics.Meter] = None
            FloTelemetry._initialized = True

    def configure(
        self,
        service_name: str = 'flo_ai',
        service_version: str = '1.0.0',
        environment: str = 'development',
        otlp_endpoint: Optional[str] = None,
        console_export: bool = False,
        additional_attributes: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Configure OpenTelemetry for the flo_ai framework.

        Args:
            service_name: Name of the service for telemetry
            service_version: Version of the service
            environment: Environment (development, production, etc.)
            otlp_endpoint: OTLP endpoint for exporting telemetry (e.g., http://localhost:4317)
            console_export: Whether to export to console for debugging
            additional_attributes: Additional resource attributes
        """
        # Create resource with service information
        resource_attrs = {
            'service.name': service_name,
            'service.version': service_version,
            'deployment.environment': environment,
        }

        if additional_attributes:
            resource_attrs.update(additional_attributes)

        resource = Resource.create(resource_attrs)

        # Configure tracing
        self.tracer_provider = TracerProvider(resource=resource)

        # Add span processors
        if console_export:
            console_processor = BatchSpanProcessor(ConsoleSpanExporter())
            self.tracer_provider.add_span_processor(console_processor)

        # Add OTLP exporter if endpoint is provided
        if otlp_endpoint:
            otlp_exporter = OTLPSpanExporter(endpoint=otlp_endpoint, insecure=True)
            otlp_processor = BatchSpanProcessor(otlp_exporter)
            self.tracer_provider.add_span_processor(otlp_processor)

        # Set the tracer provider
        trace.set_tracer_provider(self.tracer_provider)
        self.tracer = trace.get_tracer(__name__)

        # Configure metrics
        metric_readers = []

        if console_export:
            console_reader = PeriodicExportingMetricReader(
                ConsoleMetricExporter(), export_interval_millis=5000
            )
            metric_readers.append(console_reader)

        if otlp_endpoint:
            otlp_metric_exporter = OTLPMetricExporter(
                endpoint=otlp_endpoint, insecure=True
            )
            otlp_reader = PeriodicExportingMetricReader(
                otlp_metric_exporter, export_interval_millis=5000
            )
            metric_readers.append(otlp_reader)

        if metric_readers:
            self.meter_provider = MeterProvider(
                resource=resource,
                metric_readers=metric_readers,
            )
            metrics.set_meter_provider(self.meter_provider)
            self.meter = metrics.get_meter(__name__)

    def get_tracer(self) -> Optional[trace.Tracer]:
        """Get the configured tracer instance"""
        return self.tracer

    def get_meter(self) -> Optional[metrics.Meter]:
        """Get the configured meter instance"""
        return self.meter

    def shutdown(self) -> None:
        """Shutdown telemetry providers and flush data"""
        if self.tracer_provider:
            self.tracer_provider.shutdown()
        if self.meter_provider:
            self.meter_provider.shutdown()


# Global telemetry instance
_global_telemetry = FloTelemetry()


def configure_telemetry(
    service_name: str = 'flo_ai',
    service_version: str = '1.0.0',
    environment: str = None,
    otlp_endpoint: str = None,
    console_export: bool = False,
    additional_attributes: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Configure OpenTelemetry for flo_ai.

    Args:
        service_name: Name of the service
        service_version: Version of the service
        environment: Environment (defaults to FLO_ENV or 'development')
        otlp_endpoint: OTLP endpoint (defaults to FLO_OTLP_ENDPOINT or None)
        console_export: Export to console for debugging
        additional_attributes: Additional resource attributes

    Example:
        >>> from flo_ai.telemetry import configure_telemetry
        >>> configure_telemetry(
        ...     service_name="my_ai_app",
        ...     otlp_endpoint="http://localhost:4317",
        ...     console_export=True
        ... )
    """
    # Use environment variables as defaults
    if environment is None:
        environment = os.getenv('FLO_ENV', 'development')

    if otlp_endpoint is None:
        otlp_endpoint = os.getenv('FLO_OTLP_ENDPOINT')

    _global_telemetry.configure(
        service_name=service_name,
        service_version=service_version,
        environment=environment,
        otlp_endpoint=otlp_endpoint,
        console_export=console_export,
        additional_attributes=additional_attributes,
    )


def get_tracer() -> Optional[trace.Tracer]:
    """
    Get the global tracer instance.

    Returns:
        Tracer instance or None if not configured

    Example:
        >>> from flo_ai.telemetry import get_tracer
        >>> tracer = get_tracer()
        >>> if tracer:
        ...     with tracer.start_as_current_span("my_operation"):
        ...         # Your code here
        ...         pass
    """
    return _global_telemetry.get_tracer()


def get_meter() -> Optional[metrics.Meter]:
    """
    Get the global meter instance.

    Returns:
        Meter instance or None if not configured

    Example:
        >>> from flo_ai.telemetry import get_meter
        >>> meter = get_meter()
        >>> if meter:
        ...     counter = meter.create_counter("operations_total")
        ...     counter.add(1, {"operation": "inference"})
    """
    return _global_telemetry.get_meter()


def shutdown_telemetry() -> None:
    """
    Shutdown telemetry and flush all data.

    Call this at the end of your application lifecycle.

    Example:
        >>> from flo_ai.telemetry import shutdown_telemetry
        >>> shutdown_telemetry()
    """
    _global_telemetry.shutdown()



from typing import Any, AsyncIterator, Dict, List, Optional
from .openai_llm import OpenAI
from flo_ai.telemetry.instrumentation import trace_llm_stream


class OpenAIVLLM(OpenAI):
    def __init__(
        self,
        base_url: str,
        model: str,
        api_key: str = None,
        temperature: float = 0.7,
        **kwargs,
    ):
        super().__init__(
            model=model,
            api_key=api_key,
            temperature=temperature,
            base_url=base_url,
            **kwargs,
        )

        # Store base_url attribute
        self.base_url = base_url

    # overriden
    async def generate(
        self, messages: list[dict], output_schema: dict = None, **kwargs
    ) -> Any:
        # Convert output_schema to OpenAI format if provided
        if output_schema:
            kwargs['response_format'] = {
                'type': 'json_schema',
                'json_schema': {
                    'name': output_schema.get('title', 'default'),
                    'schema': output_schema.get('schema', output_schema),
                },
            }

            # Add JSON format instruction to the system prompt
            if messages and messages[0]['role'] == 'system':
                messages[0]['content'] = (
                    messages[0]['content']
                    + f'\n\nPlease provide your response in JSON format according to the specified schema. \n\n {output_schema}'
                )
            else:
                messages.insert(
                    0,
                    {
                        'role': 'system',
                        'content': f'Please provide your response in JSON format according to the specified schema.\n \n {output_schema}',
                    },
                )

        # Prepare OpenAI API parameters
        vllm_openai_kwargs = {
            'model': self.model,
            'messages': messages,
            'temperature': self.temperature,
            **kwargs,
            **self.kwargs,
        }

        # Make the API call
        response = await self.client.chat.completions.create(**vllm_openai_kwargs)
        message = response.choices[0].message

        # Return the full message object instead of just the content
        return message

    @trace_llm_stream(provider='openai_vllm')
    async def stream(
        self,
        messages: List[Dict[str, Any]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from vLLM-hosted OpenAI-compatible endpoint."""
        vllm_openai_kwargs = {
            'model': self.model,
            'messages': messages,
            'temperature': self.temperature,
            'stream': True,
            **kwargs,
            **self.kwargs,
        }

        if functions:
            vllm_openai_kwargs['functions'] = functions
        response = await self.client.chat.completions.create(**vllm_openai_kwargs)
        async for chunk in response:
            choices = getattr(chunk, 'choices', []) or []
            for choice in choices:
                delta = getattr(choice, 'delta', None)
                if delta is None:
                    continue
                content = getattr(delta, 'content', None)
                if content:
                    yield {'content': content}



"""
Event system for Arium workflow execution monitoring.

This module provides event types and data structures for tracking workflow execution,
including node starts/completions, router decisions, and workflow lifecycle events.
"""

from enum import Enum
from dataclasses import dataclass
from typing import Optional
import time
from flo_ai.utils.logger import logger


class AriumEventType(Enum):
    """Enumeration of all possible Arium workflow events."""

    WORKFLOW_STARTED = 'workflow_started'
    WORKFLOW_COMPLETED = 'workflow_completed'
    WORKFLOW_FAILED = 'workflow_failed'
    NODE_STARTED = 'node_started'
    NODE_COMPLETED = 'node_completed'
    NODE_FAILED = 'node_failed'
    ROUTER_DECISION = 'router_decision'
    EDGE_TRAVERSED = 'edge_traversed'


@dataclass
class AriumEvent:
    """
    Data structure representing a single workflow execution event.

    Attributes:
        event_type: The type of event that occurred
        timestamp: Unix timestamp when the event occurred
        node_name: Name of the node involved (if applicable)
        node_type: Type of node ('agent', 'tool', 'start', 'end')
        execution_time: Time taken for node execution in seconds
        error: Error message if the event represents a failure
        router_choice: The node chosen by a router decision
        metadata: Additional event-specific data
    """

    event_type: AriumEventType
    timestamp: float
    node_name: Optional[str] = None
    node_type: Optional[str] = None
    execution_time: Optional[float] = None
    error: Optional[str] = None
    router_choice: Optional[str] = None
    metadata: Optional[dict] = None


def default_event_callback(event: AriumEvent) -> None:
    """
    Default callback function that prints workflow events to console with formatting.

    This provides useful output for debugging and monitoring workflow execution
    without requiring any custom callback setup.

    Args:
        event: The AriumEvent to process and display
    """
    timestamp = time.strftime('%H:%M:%S', time.localtime(event.timestamp))

    if event.event_type == AriumEventType.WORKFLOW_STARTED:
        logger.info(f'🚀 [{timestamp}] Workflow started')

    elif event.event_type == AriumEventType.WORKFLOW_COMPLETED:
        logger.info(f'✅ [{timestamp}] Workflow completed')

    elif event.event_type == AriumEventType.WORKFLOW_FAILED:
        logger.error(f'❌ [{timestamp}] Workflow failed: {event.error}')

    elif event.event_type == AriumEventType.NODE_STARTED:
        node_desc = (
            f'{event.node_type}: {event.node_name}'
            if event.node_type
            else event.node_name
        )
        logger.info(f'⚡ [{timestamp}] Started {node_desc}')

    elif event.event_type == AriumEventType.NODE_COMPLETED:
        duration = f' ({event.execution_time:.2f}s)' if event.execution_time else ''
        logger.info(f'✅ [{timestamp}] Completed {event.node_name}{duration}')

    elif event.event_type == AriumEventType.NODE_FAILED:
        logger.error(f'❌ [{timestamp}] Failed {event.node_name}: {event.error}')

    elif event.event_type == AriumEventType.ROUTER_DECISION:
        logger.info(f'🔀 [{timestamp}] Router chose: {event.router_choice}')

    elif event.event_type == AriumEventType.EDGE_TRAVERSED:
        logger.info(f'➡️  [{timestamp}] Moving from {event.node_name} to next node')



"""
Instrumentation utilities for flo_ai components
"""

from typing import Optional, Dict, Any, Callable
from functools import wraps
from opentelemetry.trace import Status, StatusCode, Span
from .telemetry import get_tracer, get_meter
import time
import asyncio


class LLMMetrics:
    """Metrics for LLM operations"""

    def __init__(self):
        self.meter = get_meter()
        if self.meter:
            # Token counters
            self.token_counter = self.meter.create_counter(
                name='llm.tokens.total',
                description='Total number of tokens used',
                unit='tokens',
            )

            self.prompt_tokens_counter = self.meter.create_counter(
                name='llm.tokens.prompt',
                description='Number of prompt tokens',
                unit='tokens',
            )

            self.completion_tokens_counter = self.meter.create_counter(
                name='llm.tokens.completion',
                description='Number of completion tokens',
                unit='tokens',
            )

            # Request counters
            self.request_counter = self.meter.create_counter(
                name='llm.requests.total',
                description='Total number of LLM requests',
                unit='requests',
            )

            self.error_counter = self.meter.create_counter(
                name='llm.errors.total',
                description='Total number of LLM errors',
                unit='errors',
            )

            # Latency histogram
            self.latency_histogram = self.meter.create_histogram(
                name='llm.request.duration',
                description='Duration of LLM requests',
                unit='ms',
            )

            # Streaming metrics
            self.stream_counter = self.meter.create_counter(
                name='llm.streams.total',
                description='Total number of LLM stream requests',
                unit='streams',
            )

            self.stream_chunks_counter = self.meter.create_counter(
                name='llm.stream.chunks.total',
                description='Total number of stream chunks received',
                unit='chunks',
            )

            self.stream_duration_histogram = self.meter.create_histogram(
                name='llm.stream.duration',
                description='Duration of LLM streaming requests',
                unit='ms',
            )

    def record_tokens(
        self,
        total_tokens: int = 0,
        prompt_tokens: int = 0,
        completion_tokens: int = 0,
        model: str = '',
        provider: str = '',
    ):
        """Record token usage"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider}

        if total_tokens > 0:
            self.token_counter.add(total_tokens, attributes)
        if prompt_tokens > 0:
            self.prompt_tokens_counter.add(prompt_tokens, attributes)
        if completion_tokens > 0:
            self.completion_tokens_counter.add(completion_tokens, attributes)

    def record_request(
        self, model: str = '', provider: str = '', status: str = 'success'
    ):
        """Record LLM request"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider, 'status': status}
        self.request_counter.add(1, attributes)

    def record_error(self, model: str = '', provider: str = '', error_type: str = ''):
        """Record LLM error"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider, 'error_type': error_type}
        self.error_counter.add(1, attributes)

    def record_latency(self, duration_ms: float, model: str = '', provider: str = ''):
        """Record request latency"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider}
        self.latency_histogram.record(duration_ms, attributes)

    def record_stream(
        self, model: str = '', provider: str = '', status: str = 'success'
    ):
        """Record LLM stream request"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider, 'status': status}
        self.stream_counter.add(1, attributes)

    def record_stream_chunks(
        self, chunk_count: int, model: str = '', provider: str = ''
    ):
        """Record stream chunks received"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider}
        self.stream_chunks_counter.add(chunk_count, attributes)

    def record_stream_latency(
        self, duration_ms: float, model: str = '', provider: str = ''
    ):
        """Record stream request latency"""
        if not self.meter:
            return

        attributes = {'model': model, 'provider': provider}
        self.stream_duration_histogram.record(duration_ms, attributes)


class AgentMetrics:
    """Metrics for Agent operations"""

    def __init__(self):
        self.meter = get_meter()
        if self.meter:
            # Execution counters
            self.execution_counter = self.meter.create_counter(
                name='agent.executions.total',
                description='Total number of agent executions',
                unit='executions',
            )

            self.tool_call_counter = self.meter.create_counter(
                name='agent.tool_calls.total',
                description='Total number of tool calls',
                unit='calls',
            )

            self.retry_counter = self.meter.create_counter(
                name='agent.retries.total',
                description='Total number of retries',
                unit='retries',
            )

            self.error_counter = self.meter.create_counter(
                name='agent.errors.total',
                description='Total number of agent errors',
                unit='errors',
            )

            # Latency histogram
            self.latency_histogram = self.meter.create_histogram(
                name='agent.execution.duration',
                description='Duration of agent executions',
                unit='ms',
            )

    def record_execution(
        self, agent_name: str = '', agent_type: str = '', status: str = 'success'
    ):
        """Record agent execution"""
        if not self.meter:
            return

        attributes = {
            'agent_name': agent_name,
            'agent_type': agent_type,
            'status': status,
        }
        self.execution_counter.add(1, attributes)

    def record_tool_call(
        self, agent_name: str = '', tool_name: str = '', status: str = 'success'
    ):
        """Record tool call"""
        if not self.meter:
            return

        attributes = {
            'agent_name': agent_name,
            'tool_name': tool_name,
            'status': status,
        }
        self.tool_call_counter.add(1, attributes)

    def record_retry(self, agent_name: str = '', reason: str = ''):
        """Record retry attempt"""
        if not self.meter:
            return

        attributes = {'agent_name': agent_name, 'reason': reason}
        self.retry_counter.add(1, attributes)

    def record_error(self, agent_name: str = '', error_type: str = ''):
        """Record agent error"""
        if not self.meter:
            return

        attributes = {'agent_name': agent_name, 'error_type': error_type}
        self.error_counter.add(1, attributes)

    def record_latency(
        self, duration_ms: float, agent_name: str = '', agent_type: str = ''
    ):
        """Record execution latency"""
        if not self.meter:
            return

        attributes = {'agent_name': agent_name, 'agent_type': agent_type}
        self.latency_histogram.record(duration_ms, attributes)


class WorkflowMetrics:
    """Metrics for Arium workflow operations"""

    def __init__(self):
        self.meter = get_meter()
        if self.meter:
            # Workflow counters
            self.workflow_counter = self.meter.create_counter(
                name='workflow.executions.total',
                description='Total number of workflow executions',
                unit='executions',
            )

            self.node_counter = self.meter.create_counter(
                name='workflow.nodes.executed',
                description='Total number of nodes executed',
                unit='nodes',
            )

            self.error_counter = self.meter.create_counter(
                name='workflow.errors.total',
                description='Total number of workflow errors',
                unit='errors',
            )

            # Latency histograms
            self.workflow_latency = self.meter.create_histogram(
                name='workflow.execution.duration',
                description='Duration of workflow executions',
                unit='ms',
            )

            self.node_latency = self.meter.create_histogram(
                name='workflow.node.duration',
                description='Duration of node executions',
                unit='ms',
            )

    def record_workflow(self, workflow_name: str = '', status: str = 'success'):
        """Record workflow execution"""
        if not self.meter:
            return

        attributes = {'workflow_name': workflow_name, 'status': status}
        self.workflow_counter.add(1, attributes)

    def record_node(
        self,
        workflow_name: str = '',
        node_name: str = '',
        node_type: str = '',
        status: str = 'success',
    ):
        """Record node execution"""
        if not self.meter:
            return

        attributes = {
            'workflow_name': workflow_name,
            'node_name': node_name,
            'node_type': node_type,
            'status': status,
        }
        self.node_counter.add(1, attributes)

    def record_error(self, workflow_name: str = '', error_type: str = ''):
        """Record workflow error"""
        if not self.meter:
            return

        attributes = {'workflow_name': workflow_name, 'error_type': error_type}
        self.error_counter.add(1, attributes)

    def record_workflow_latency(self, duration_ms: float, workflow_name: str = ''):
        """Record workflow latency"""
        if not self.meter:
            return

        attributes = {'workflow_name': workflow_name}
        self.workflow_latency.record(duration_ms, attributes)

    def record_node_latency(
        self,
        duration_ms: float,
        workflow_name: str = '',
        node_name: str = '',
        node_type: str = '',
    ):
        """Record node latency"""
        if not self.meter:
            return

        attributes = {
            'workflow_name': workflow_name,
            'node_name': node_name,
            'node_type': node_type,
        }
        self.node_latency.record(duration_ms, attributes)


# Global metric instances
llm_metrics = LLMMetrics()
agent_metrics = AgentMetrics()
workflow_metrics = WorkflowMetrics()


def trace_llm_call(provider: str = '', model: str = ''):
    """
    Decorator to trace LLM API calls

    Args:
        provider: LLM provider name (e.g., 'openai', 'anthropic', 'gemini')
        model: Model name

    Example:
        @trace_llm_call(provider="openai", model="gpt-4")
        async def generate(self, messages):
            ...
    """

    def decorator(func: Callable):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            tracer = get_tracer()
            if not tracer:
                return await func(*args, **kwargs)

            # Extract self to get instance attributes
            self_arg = args[0] if args else None
            actual_model = model or (getattr(self_arg, 'model', '') if self_arg else '')
            actual_provider = provider or (
                self_arg.__class__.__name__ if self_arg else ''
            )

            with tracer.start_as_current_span(
                f'llm.{actual_provider}.generate',
                attributes={
                    'llm.provider': actual_provider,
                    'llm.model': actual_model,
                    'llm.temperature': getattr(self_arg, 'temperature', 0.0)
                    if self_arg
                    else 0.0,
                },
            ) as span:
                start_time = time.time()
                try:
                    result = await func(*args, **kwargs)

                    # Record success
                    duration_ms = (time.time() - start_time) * 1000
                    llm_metrics.record_request(actual_model, actual_provider, 'success')
                    llm_metrics.record_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.OK))
                    span.set_attribute('llm.response.received', True)

                    return result

                except Exception as e:
                    # Record error
                    duration_ms = (time.time() - start_time) * 1000
                    error_type = type(e).__name__

                    llm_metrics.record_request(actual_model, actual_provider, 'error')
                    llm_metrics.record_error(actual_model, actual_provider, error_type)
                    llm_metrics.record_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.set_attribute('error.type', error_type)
                    span.set_attribute('error.message', str(e))

                    raise

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            tracer = get_tracer()
            if not tracer:
                return func(*args, **kwargs)

            self_arg = args[0] if args else None
            actual_model = model or (getattr(self_arg, 'model', '') if self_arg else '')
            actual_provider = provider or (
                self_arg.__class__.__name__ if self_arg else ''
            )

            with tracer.start_as_current_span(
                f'llm.{actual_provider}.generate',
                attributes={
                    'llm.provider': actual_provider,
                    'llm.model': actual_model,
                },
            ) as span:
                start_time = time.time()
                try:
                    result = func(*args, **kwargs)

                    duration_ms = (time.time() - start_time) * 1000
                    llm_metrics.record_request(actual_model, actual_provider, 'success')
                    llm_metrics.record_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.OK))
                    return result

                except Exception as e:
                    duration_ms = (time.time() - start_time) * 1000
                    error_type = type(e).__name__

                    llm_metrics.record_request(actual_model, actual_provider, 'error')
                    llm_metrics.record_error(actual_model, actual_provider, error_type)
                    llm_metrics.record_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.set_attribute('error.type', error_type)

                    raise

        # Return appropriate wrapper based on function type
        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def trace_llm_stream(provider: str = '', model: str = ''):
    """
    Decorator to trace LLM streaming API calls

    Args:
        provider: LLM provider name (e.g., 'openai', 'anthropic', 'gemini')
        model: Model name

    Example:
        @trace_llm_stream(provider="openai", model="gpt-4")
        async def stream(self, messages):
            ...
    """

    def decorator(func: Callable):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            tracer = get_tracer()
            if not tracer:
                async for chunk in func(*args, **kwargs):
                    yield chunk
                return

            # Extract self to get instance attributes
            self_arg = args[0] if args else None
            actual_model = model or (getattr(self_arg, 'model', '') if self_arg else '')
            actual_provider = provider or (
                self_arg.__class__.__name__ if self_arg else ''
            )

            with tracer.start_as_current_span(
                f'llm.{actual_provider}.stream',
                attributes={
                    'llm.provider': actual_provider,
                    'llm.model': actual_model,
                    'llm.temperature': getattr(self_arg, 'temperature', 0.0)
                    if self_arg
                    else 0.0,
                    'llm.operation': 'stream',
                },
            ) as span:
                start_time = time.time()
                chunk_count = 0
                try:
                    # Record stream start
                    llm_metrics.record_stream(actual_model, actual_provider, 'start')

                    # Track the streaming response
                    async for chunk in func(*args, **kwargs):
                        chunk_count += 1
                        yield chunk

                    # Record success
                    duration_ms = (time.time() - start_time) * 1000
                    llm_metrics.record_stream(actual_model, actual_provider, 'success')
                    llm_metrics.record_stream_chunks(
                        chunk_count, actual_model, actual_provider
                    )
                    llm_metrics.record_stream_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.OK))
                    span.set_attribute('llm.stream.chunks', chunk_count)
                    span.set_attribute('llm.stream.duration_ms', duration_ms)
                    span.set_attribute('llm.stream.completed', True)

                except Exception as e:
                    # Record error
                    duration_ms = (time.time() - start_time) * 1000
                    error_type = type(e).__name__

                    llm_metrics.record_stream(actual_model, actual_provider, 'error')
                    llm_metrics.record_error(actual_model, actual_provider, error_type)
                    llm_metrics.record_stream_latency(
                        duration_ms, actual_model, actual_provider
                    )

                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.set_attribute('error.type', error_type)
                    span.set_attribute('error.message', str(e))
                    span.set_attribute('llm.stream.chunks', chunk_count)
                    span.set_attribute('llm.stream.duration_ms', duration_ms)

                    raise

        return async_wrapper

    return decorator


def trace_agent_execution(agent_name: str = ''):
    """
    Decorator to trace agent executions

    Args:
        agent_name: Name of the agent

    Example:
        @trace_agent_execution(agent_name="research_agent")
        async def run(self, inputs):
            ...
    """

    def decorator(func: Callable):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            tracer = get_tracer()
            if not tracer:
                return await func(*args, **kwargs)

            self_arg = args[0] if args else None
            actual_agent_name = agent_name or (
                getattr(self_arg, 'name', '') if self_arg else ''
            )
            agent_type = getattr(self_arg, 'agent_type', '') if self_arg else ''

            with tracer.start_as_current_span(
                f'agent.{actual_agent_name}.run',
                attributes={
                    'agent.name': actual_agent_name,
                    'agent.type': str(agent_type),
                },
            ) as span:
                start_time = time.time()
                try:
                    result = await func(*args, **kwargs)

                    duration_ms = (time.time() - start_time) * 1000
                    agent_metrics.record_execution(
                        actual_agent_name, str(agent_type), 'success'
                    )
                    agent_metrics.record_latency(
                        duration_ms, actual_agent_name, str(agent_type)
                    )

                    span.set_status(Status(StatusCode.OK))
                    span.set_attribute(
                        'agent.result.length', len(str(result)) if result else 0
                    )

                    return result

                except Exception as e:
                    duration_ms = (time.time() - start_time) * 1000
                    error_type = type(e).__name__

                    agent_metrics.record_execution(
                        actual_agent_name, str(agent_type), 'error'
                    )
                    agent_metrics.record_error(actual_agent_name, error_type)
                    agent_metrics.record_latency(
                        duration_ms, actual_agent_name, str(agent_type)
                    )

                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.set_attribute('error.type', error_type)

                    raise

        return async_wrapper

    return decorator


def add_span_attributes(span: Optional[Span], attributes: Dict[str, Any]) -> None:
    """
    Helper to add attributes to a span safely

    Args:
        span: OpenTelemetry span
        attributes: Dictionary of attributes to add
    """
    if span and attributes:
        for key, value in attributes.items():
            # OpenTelemetry only supports certain types
            if isinstance(value, (str, bool, int, float)):
                span.set_attribute(key, value)
            else:
                span.set_attribute(key, str(value))



import json
import csv
import yaml
from io import StringIO
from typing import List, Dict, Any, Optional, Literal
from pydantic import BaseModel, Field, create_model
from dataclasses import dataclass


@dataclass
class ParseContract:
    name: str
    fields: List[Dict[str, Any]]


class FloJsonParser:
    def __init__(self, parse_contract: ParseContract):
        self.contract = parse_contract
        self._cached_models = {}
        super().__init__()

    def __dict_list_to_csv_string(self, data):
        if not data or len(data) == 0:
            return '```No data provided```'
        headers = data[0].keys()
        output = StringIO()

        writer = csv.DictWriter(output, fieldnames=headers)
        writer.writeheader()
        writer.writerows(data)

        csv_string = output.getvalue()
        output.close()

        return f'```\n{csv_string}```'

    def __create_nested_model(
        self, field_def: Dict[str, Any], model_name: str
    ) -> BaseModel:
        """Creates a nested Pydantic model for object types"""
        if model_name in self._cached_models:
            return self._cached_models[model_name]

        nested_fields = {}
        for nested_field in field_def['fields']:
            nested_type = self.__get_field_type_annotation(
                nested_field, f"{model_name}_{nested_field['name']}"
            )
            field_description = nested_field['description']
            nested_fields[nested_field['name']] = (
                nested_type,
                Field(..., description=field_description),
            )

        NestedModel = create_model(model_name, **nested_fields)
        self._cached_models[model_name] = NestedModel
        return NestedModel

    def __get_field_type_annotation(
        self, field: Dict[str, Any], model_name: str
    ) -> Any:
        """Determines the type annotation for a field, handling nested objects"""
        type_mapping = {
            'str': str,
            'int': int,
            'bool': bool,
            'float': float,
            'literal': self.__create_literal_type,
            'object': lambda f: self.__create_nested_model(f, model_name),
            'array': lambda f: List[
                self.__get_field_type_annotation(f['items'], f'{model_name}_item')
            ],
        }

        field_type = field['type']
        type_handler = type_mapping.get(field_type)

        if type_handler is None:
            raise ValueError(f'Unsupported type: {field_type}')

        return (
            type_handler(field)
            if field_type in ['literal', 'object', 'array']
            else type_handler
        )

    def __create_literal_type(self, field: Dict[str, Any]) -> Any:
        """Creates a Literal type from field definition"""
        literal_values = field.get('values', [])
        if not literal_values:
            raise ValueError(
                f"Field '{field['name']}' of type 'literal' must specify 'values'."
            )
        literals = [literal_value['value'] for literal_value in literal_values]
        return Literal[tuple(literals)]

    def get_format(self) -> BaseModel:
        return self.__create_contract_from_json()

    def __create_contract_from_json(self) -> BaseModel:
        pydantic_fields = {}
        for field in self.contract.fields:
            field_type = self.__get_field_type_annotation(
                field, f"{self.contract.name}_{field['name']}"
            )

            if field['type'] == 'literal':
                literal_values = field.get('values', [])
                default_prompt = field.get('default_value_prompt', '')
                field_description = f"""
                {field['description']}
                Following are the list of possibles values and its correponding description:
                {self.__dict_list_to_csv_string(literal_values)}

                This should be one of the values in the `value` column in the above csv.
                {default_prompt}
                """
            else:
                field_description = field['description']

            pydantic_fields[field['name']] = (
                field_type,
                Field(..., description=field_description),
            )

        DynamicModel = create_model(self.contract.name, **pydantic_fields)
        return DynamicModel

    @staticmethod
    def create(json_dict: Optional[Dict] = None, json_path: Optional[str] = None):
        return FloJsonParser.Builder(json_dict=json_dict, json_path=json_path).build()

    class Builder:
        def __init__(
            self, json_dict: Optional[Dict] = None, json_path: Optional[str] = None
        ):
            if json_dict is None and json_path is None:
                raise ValueError(
                    'Either of json_dict or json_path is required to build a FloJsonParser'
                )
            self.json_dict = json_dict
            self.json_path = json_path

        def build(self):
            if self.json_dict:
                name = self.json_dict['name']
                fields = self.json_dict['fields']
            else:
                with open(self.json_path) as f:
                    json_contract = json.load(f)
                name = json_contract['name']
                fields = json_contract['fields']
            return FloJsonParser(ParseContract(name=name, fields=fields))


class FloYamlParser(FloJsonParser):
    """
    A parser class that handles YAML-based parser definitions for Flo agents.
    Extends FloJsonParser to reuse the model creation logic while adding YAML-specific functionality.
    """

    @staticmethod
    def create(yaml_dict: Optional[Dict] = None, yaml_path: Optional[str] = None):
        """
        Create a FloYamlParser instance from either a YAML dictionary or a YAML file path.

        Args:
            yaml_dict: A dictionary containing the YAML parser definition
            yaml_path: Path to a YAML file containing the parser definition

        Returns:
            FloYamlParser: A configured parser instance
        """
        return FloYamlParser.Builder(yaml_dict=yaml_dict, yaml_path=yaml_path).build()

    class Builder:
        def __init__(
            self, yaml_dict: Optional[Dict] = None, yaml_path: Optional[str] = None
        ):
            if yaml_dict is None and yaml_path is None:
                raise ValueError(
                    'Either yaml_dict or yaml_path is required to build a FloYamlParser'
                )
            self.yaml_dict = yaml_dict
            self.yaml_path = yaml_path

        def build(self):
            if self.yaml_dict:
                parser_def = self.yaml_dict
            else:
                with open(self.yaml_path) as f:
                    parser_def = yaml.safe_load(f)

            # Extract parser definition from agent YAML
            if 'agent' in parser_def and 'parser' in parser_def['agent']:
                parser_def = parser_def['agent']['parser']

            # Extract required fields
            name = parser_def['name']
            fields = parser_def['fields']

            # Process fields to handle examples and required flag
            processed_fields = []
            for field in fields:
                processed_field = field.copy()

                # Handle examples in literal values
                if field['type'] == 'literal' and 'values' in field:
                    for value in field['values']:
                        if 'examples' in value:
                            # Add examples to description
                            examples_str = '\nExamples:\n' + '\n'.join(
                                f'- {ex}' for ex in value['examples']
                            )
                            value['description'] = value['description'] + examples_str
                            del value['examples']

                # Remove required flag as it's not used in model creation
                if 'required' in processed_field:
                    del processed_field['required']

                processed_fields.append(processed_field)

            return FloYamlParser(ParseContract(name=name, fields=processed_fields))



import base64
import asyncio
from typing import Dict, Any, List, Optional, AsyncIterator
from .base_llm import BaseLLM
from flo_ai.models.chat_message import ImageMessageContent
from google import genai
from google.genai import types
from flo_ai.tool.base_tool import Tool
from flo_ai.telemetry.instrumentation import (
    trace_llm_call,
    trace_llm_stream,
    llm_metrics,
    add_span_attributes,
)
from flo_ai.telemetry import get_tracer
from opentelemetry import trace


class Gemini(BaseLLM):
    def __init__(
        self,
        model: str = 'gemini-2.5-flash',
        temperature: float = 0.7,
        api_key: Optional[str] = None,
        base_url: str = None,
        custom_headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        super().__init__(model, api_key, temperature, **kwargs)
        # Configure http_options for proxy or custom base_url
        http_options = {'base_url': base_url} if base_url else {}
        if base_url and self.api_key:
            # For custom base_url (proxy), set Authorization header explicitly
            http_options['headers'] = {'Authorization': f'Bearer {self.api_key}'}
            # Merge custom headers if provided (proxy scenario)
            if custom_headers:
                http_options['headers'].update(custom_headers)

        # Initialize client based on configuration
        if http_options:
            self.client = genai.Client(http_options=http_options)
        elif self.api_key:
            self.client = genai.Client(api_key=self.api_key)
        else:
            self.client = genai.Client()

    @trace_llm_call(provider='gemini')
    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        # Convert messages to Gemini format
        contents = []
        system_prompt = ''

        for msg in messages:
            role = msg['role']
            message_content = msg['content']

            if role == 'system':
                system_prompt += f'{message_content}\n'
            else:
                contents.append(message_content)

        try:
            # Prepare generation config
            # Merge instance kwargs with method kwargs
            config_kwargs = {**self.kwargs, **kwargs}
            generation_config = types.GenerateContentConfig(
                temperature=self.temperature,
                system_instruction=system_prompt,
                **config_kwargs,
            )

            # Add tools if functions are provided
            if functions:
                tools = types.Tool(function_declarations=functions)
                generation_config.tools = [tools]

            # Add structured output configuration if output_schema is provided
            if output_schema:
                generation_config.response_mime_type = 'application/json'
                generation_config.response_schema = output_schema

            # Make the API call (run in thread pool to avoid blocking event loop)
            response = await asyncio.to_thread(
                self.client.models.generate_content,
                model=self.model,
                contents=contents,
                config=generation_config,
            )

            # Record token usage if available
            if hasattr(response, 'usage_metadata') and response.usage_metadata:
                usage = response.usage_metadata
                prompt_tokens = getattr(usage, 'prompt_token_count', 0)
                completion_tokens = getattr(usage, 'candidates_token_count', 0)
                total_tokens = getattr(usage, 'total_token_count', 0)

                llm_metrics.record_tokens(
                    total_tokens=total_tokens,
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    model=self.model,
                    provider='gemini',
                )

                # Add token info to current span
                tracer = get_tracer()
                if tracer:
                    current_span = trace.get_current_span()
                    add_span_attributes(
                        current_span,
                        {
                            'llm.tokens.prompt': prompt_tokens,
                            'llm.tokens.completion': completion_tokens,
                            'llm.tokens.total': total_tokens,
                        },
                    )

            # Check for function call in the response
            if (
                functions
                and response.candidates
                and response.candidates[0].content.parts
            ):
                part = response.candidates[0].content.parts[0]
                if hasattr(part, 'function_call') and part.function_call:
                    function_call = part.function_call
                    return {
                        'content': response.text,
                        'function_call': {
                            'name': function_call.name,
                            'arguments': function_call.args,
                        },
                    }

            # Return regular text response
            response_text = (
                response.text if hasattr(response, 'text') else str(response)
            )
            return {'content': response_text}

        except Exception as e:
            raise Exception(f'Error in Gemini API call: {str(e)}')

    @trace_llm_stream(provider='gemini')
    async def stream(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from Gemini as they are generated"""
        # Convert messages to Gemini format
        contents = []
        system_prompt = ''

        for msg in messages:
            role = msg['role']
            message_content = msg['content']

            if role == 'system':
                system_prompt += f'{message_content}\n'
            else:
                contents.append(message_content)

        # Prepare generation config
        # Merge instance kwargs with method kwargs
        config_kwargs = {**self.kwargs, **kwargs}
        generation_config = types.GenerateContentConfig(
            temperature=self.temperature,
            system_instruction=system_prompt,
            **config_kwargs,
        )

        # Add tools if functions are provided
        if functions:
            tools = types.Tool(function_declarations=functions)
            generation_config.tools = [tools]

        # Get stream in thread to avoid blocking the initial call
        stream = await asyncio.to_thread(
            self.client.models.generate_content_stream,
            model=self.model,
            contents=contents,
            config=generation_config,
        )

        # Helper to get next chunk in thread pool
        def get_next_chunk():
            try:
                return next(stream)
            except StopIteration:
                return None

        # Iterate over synchronous stream without blocking event loop
        while True:
            chunk = await asyncio.to_thread(get_next_chunk)
            if chunk is None:
                break
            if hasattr(chunk, 'text') and chunk.text:
                yield {'content': chunk.text}

    def get_message_content(self, response: Any) -> str:
        """Extract message content from response"""
        if isinstance(response, dict):
            return response.get('content', '')
        return str(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a single tool for Gemini's function declarations"""
        return {
            'name': tool.name,
            'description': tool.description,
            'parameters': {
                'type': 'object',
                'properties': {
                    name: {
                        'type': info.get('type', 'string'),
                        'description': info.get('description', ''),
                        **(
                            {'items': info['items']}
                            if info.get('type') == 'array' and 'items' in info
                            else {}
                        ),
                    }
                    for name, info in tool.parameters.items()
                },
                'required': [
                    name
                    for name, info in tool.parameters.items()
                    if info.get('required', True)
                ],
            },
        }

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format tools for Gemini's function declarations"""
        return [self.format_tool_for_llm(tool) for tool in tools]

    def format_image_in_message(self, image: ImageMessageContent) -> str:
        """Format a image in the message"""

        if image.base64:
            return types.Part.from_bytes(
                data=base64.b64decode(image.base64),
                mime_type=image.mime_type,
            )
        elif image.url:
            return types.Part.from_uri(
                file_uri=image.url,
                mime_type=image.mime_type,
            )
        raise NotImplementedError(
            f'Image formatting for Gemini LLM requires either url or base64 data. Received: url={image.url}, base64={bool(image.base64)}'
        )
